#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps.Distributions.DistributionBase import *
from TransportMaps.Distributions.Inference.InferenceBase import BayesPosteriorDistribution
from TransportMaps.Distributions.FrozenDistributions import StandardNormalDistribution
from TransportMaps.Likelihoods.LikelihoodBase import IndependentLogLikelihood

__all__ = ['AR1TransitionDistribution',
           'MarkovChainDistribution',
           'SequentialHiddenMarkovChainDistribution',
           'MarkovComponentDistribution']

nax = np.newaxis

class AR1TransitionDistribution(ConditionalDistribution):
    r""" Transition probability for an auto-regressive (1) process (possibly with hyper-parameters)

    Defines the probability distribution
    :math:`\pi({\bf Z}_{k+1}\vert {\bf Z}_{k}, \Theta)=\pi({\bf Z}_{k+1} - T({\bf Z}_{k},\Theta) \vert \Theta)`
    for the auto-regressive (1) process

    .. math::

       {\bf Z}_{k+1} = T({\bf Z}_k, \Theta) + \varepsilon
       \;, \quad
       \varepsilon \sim \nu_\pi

    Args:
      pi (:class:`Distribution<TransportMaps.Distributions.ConditionalDistribution`):
        distribution :math:`\pi:\mathbb{R}^d\times\mathbb{R}^{d_\theta}\rightarrow\mathbb{R}`
      T (:class:`Map<TransportMaps.Maps.Map>`): map
        :math:`T:\mathbb{R}^{d+d_\theta}\rightarrow\mathbb{R}^d`
    """
    def __init__(self, pi, T):
        if pi.dim != T.dim_out:
            raise ValueError("The dimension of pi must match the output " + \
                             "dimension of T.")
        self._pi = pi
        if issubclass(type(pi), ConditionalDistribution):
            self._isPiCond = True
        else:
            self._isPiCond = False
        self._T = T
        self._state_dim = pi.dim
        self._hyper_dim = T.dim_in - T.dim_out
        super(AR1TransitionDistribution, self).__init__(pi.dim, T.dim_in)

    @property
    def T(self):
        return self._T

    @property
    def pi(self):
        return self._pi

    @property
    def isPiCond(self):
        return self._isPiCond

    @property
    def state_dim(self):
        return self._state_dim

    @property
    def hyper_dim(self):
        return self._hyper_dim
        
    def log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of :math:`\log\pi`
            at the ``x`` points.
        """
        frw = self.T.evaluate(y, params, idxs_slice)
        if self.isPiCond:
            return self.pi.log_pdf(x - frw, y[:,self.dim:])
        else:
            return self.pi.log_pdf(x - frw)

    def grad_x_log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla_{\bf x,y} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- values of
            :math:`\nabla_x\log\pi` at the ``x`` points.
        """
        frw = self.T.evaluate(y, params,idxs_slice)
        gx_frw = self.T.grad_x(y, params, idxs_slice)
        if self.isPiCond:
            gxlpdf = self.pi.grad_x_log_pdf(x - frw, y[:,self.dim:])
        else:
            gxlpdf = self.pi.grad_x_log_pdf(x - frw)
        out = np.zeros((x.shape[0], self.dim + self.dim_y))
        out[:,:self.dim] = gxlpdf
        out[:,self.dim:] = - np.einsum('...i,...ij->...j', gxlpdf, gx_frw)
        return out

    def hess_x_log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None)):
        r""" Evaluate :math:`\nabla^2_{\bf x,y} \log \pi({\bf x}\vert{\bf y})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          y (:class:`ndarray<numpy.ndarray>` [:math:`m,d_y`]): conditioning values
            :math:`{\bf Y}={\bf y}`
          params (dict): parameters
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) -- values of
            :math:`\nabla^2_x\log\pi` at the ``x`` points.
        """
        m = x.shape[0]
        dx = self.dim
        dy = self.dim_y
        frw = self.T.evaluate(y, params,idxs_slice)
        gx_frw = self.T.grad_x(y, params, idxs_slice)
        hx_frw = self.T.hess_x(y, params, idxs_slice)
        if self.isPiCond:
            gx_lpdf = self.pi.grad_x_log_pdf(x - frw, y[:,dx:])
            hx_lpdf = self.pi.hess_x_log_pdf(x - frw, y[:,dx:])
        else:
            gx_lpdf = self.pi.grad_x_log_pdf(x - frw)
            hx_lpdf = self.pi.hess_x_log_pdf(x - frw)
        out = np.zeros((m, dx+dy, dx+dy))
        out[:,:dx,:dx] = hx_lpdf
        out[:,:dx,dx:] = - np.einsum('...ij,...ik->...jk', hx_lpdf, gx_frw)
        out[:,dx:,:dx] = out[:,:dx,dx:].transpose((0,2,1))
        tmp = - out[:,:dx,dx:] # m x dx x dy
        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_frw)
        tmp -= np.einsum('...i,...ijk->...jk', gx_lpdf, hx_frw)
        out[:,dx:,dx:] = tmp
        return out

class MarkovChainDistribution(FactorizedDistribution):
    r""" Distribution of a Markov process (optionally with hyper-parameters)

    For the index set :math:`A=[t_0,\ldots,t_k]` with :math:`t_0<t_1<\ldots <t_k`,
    and the user defined distributions
    :math:`\pi({\bf Z}_{t_i} \vert {\bf Z}_{t_{i-1}}, \Theta)`,
    :math:`\pi({\bf Z}_{t_0} \vert \Theta)` and :math:`\pi(\Theta)`
    defines the distribution

    .. math::

       \pi(\Theta, {\bf Z}_A) = \left( \prod_{i=1}^k \pi({\bf Z}_{t_i} \vert {\bf Z}_{t_{i-1}}, \Theta) \right) \pi({\bf Z}_{t_0} \vert \Theta) \pi(\Theta)

    associated to the process :math:`{\bf Z}_A`.

    Args:
      pi_list (:class:`list` of :class:`ConditionalDistribution`):
        list of transition distributions
        :math:`\{\pi({\bf Z}_{t_0} \vert \Theta), \pi({\bf Z}_{t_1}\vert {\bf Z}_{t_{0}},\Theta), \ldots \}`
      pi_hyper (:class:`Distribution`): prior on hyper-parameters :math:`h(\Theta)`
    """
    def __init__(self, pi_list, pi_hyper=None):
        factors = []
        # Figure out dimension hyper-parameters
        if pi_hyper is None:
            hdim = 0
        else:
            hdim = pi_hyper.dim
            factors.append( (pi_hyper, list(range(hdim)), []) )
        # Check consistency
        dim = hdim
        sdim = None
        for i, pi in enumerate(pi_list):
            if i == 0:
                if issubclass(type(pi), ConditionalDistribution):
                    if pi.dim_y != hdim:
                        raise ValueError("The dimension of the %d component " % i + \
                                         "is not consistent")
                elif hdim != 0:
                    raise ValueError("The dimension of the %d component " % i + \
                                     "is not consistent")
                sdim = pi.dim
                factors.append( (pi, list(range(dim,dim+sdim)), list(range(hdim))) )
            else:
                if pi.dim != sdim or pi.dim_y != sdim + hdim:
                    raise ValueError("The dimension of the %d component " % i + \
                                     "is not consistent")
                factors.append( (pi, list(range(dim,dim+sdim)),
                                 list(range(dim-sdim,dim)) + list(range(hdim))) )
            dim += sdim
        # Init
        super(MarkovChainDistribution, self).__init__(factors)
        self.pi_list = pi_list
        self.pi_hyper = pi_hyper
        self.state_dim = sdim
        self.hyper_dim = hdim

    @property
    def nsteps(self):
        r""" Returns the number of steps (time indices) :math:`\sharp A`.
        """
        return len(self.pi_list)

    def append(self, pi):
        r""" Append a new transition distribution :math:`\pi({\bf Z}_{t_{k+1}}\vert {\bf Z}_{t_{k}},\Theta)`

        Args:
          pi (:class:`Distribution` or :class:`ConditionaDistribution`):
            transition distribution
            :math:`\pi({\bf Z}_{t_{k+1}}\vert {\bf Z}_{t_{k}},\Theta)`
        """
        # Check consistency
        if len(self.pi_list) == 0:
            if issubclass(type(pi), ConditionalDistribution):
                if pi.dim_y != self.hyper_dim:
                    raise ValueError("The dimension of the new component is not consistent")
            elif self.hyper_dim != 0:
                raise ValueError("The initial conditon prior must be conditional on " + \
                                 "the hyper parameters.")
            self.state_dim = pi.dim
            # Prepare factor
            factor = (pi, list(range(self.dim,self.dim+self.state_dim)),
                      list(range(self.hyper_dim)))
        else:
            if pi.dim != self.state_dim or pi.dim_y != self.state_dim + self.hyper_dim:
                raise ValueError("The dimension of the new component is not consistent")
            # Prepare factor
            factor = (pi, list(range(self.dim,self.dim+self.state_dim)),
                      list(range(self.dim-self.state_dim,self.dim)) + list(range(self.hyper_dim)))
        super(MarkovChainDistribution, self).append(factor)
        self.pi_list.append( pi )

class SequentialHiddenMarkovChainDistribution(BayesPosteriorDistribution):
    r""" Distribution of a sequential Hidden Markov chain model (optionally with hyper-parameters)

    For the index sets :math:`A=[t_0,\ldots,t_k]` with :math:`t_0<t_1<\ldots <t_k`, 
    :math:`B \subseteq A`, 
    the user defined transition densities (:class:`Distribution`)
    :math:`\{\pi({\bf Z}_{t_0}\vert\Theta), \pi({\bf Z}_{1}\vert{\bf Z}_{t_{0}},\Theta), \ldots \}`,
    the prior :math:`\pi(\Theta)` and 
    the log-likelihoods (:class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`)
    :math:`\{\log\mathcal{L}({\bf y}_t \vert{\bf Z}_t,\Theta)\}_{t\in B}`, defines the distribution

    .. math::

       \pi(\Theta, {\bf Z}_A \vert {\bf y}_B) =
         \left( \prod_{t\in B} \mathcal{L}({\bf y}_t \vert {\bf Z}_t, \Theta) \right)
         \left( \prod_{i=1}^k \pi({\bf Z}_{t_i}\vert{\bf Z}_{t_{i-1}},\Theta) \right)
         \pi({\bf Z}_{t_0},\Theta) \pi(\Theta)

    associated to the process :math:`{\bf Z}_A`
    
    .. note:: Each of the log-likelihoods already embed its own data :math:`{\bf y}_t`.
       The list of log-likelihoods must be of the same length of the list of
       transitions. Missing data are simulated by setting the corresponding
       entry in the list of log-likelihood to ``None``.

    Args:
      pi_list (:class:`list` of :class:`ConditionalDistribution`): list of transition densities
        :math:`[\pi({\bf Z}_{t_0}\vert\Theta), \pi({\bf Z}_{t_1}\vert{\bf Z}_{t_{0}},\Theta), \ldots ]`
      ll_list (:class:`list` of :class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
        list of log-likelihoods
        :math:`\{\log\mathcal{L}({\bf y}_t \vert {\bf Z}_t,\Theta)\}_{t\in B}`
      pi_hyper (:class:`Distribution`): prior on hyper-parameters :math:`h(\Theta)`
    """
    def __init__(self, pi_list, ll_list, pi_hyper=None):
        if len(pi_list) != len(ll_list):
            raise ValueError("Length of list of transition must be the same of " + \
                             "the list of log-likelihoods")
        prior = MarkovChainDistribution(pi_list, pi_hyper)
        # Prepare likelihood factors and check consistency
        ll_factors = []
        sdim = prior.state_dim
        hdim = prior.hyper_dim
        dim = hdim
        for i, ll in enumerate(ll_list):
            if ll is not None and ll.dim != sdim + hdim:
                raise ValueError("Dimension of log-likelihood %d is not consistent" % i)
            ll_factors.append( (ll, list(range(dim,dim+sdim)) + list(range(hdim))) )
            dim += sdim
        logL = IndependentLogLikelihood( ll_factors )
        # Init
        super(SequentialHiddenMarkovChainDistribution, self).__init__(logL, prior)

    @property
    def state_dim(self):
        return self.prior.state_dim

    @property
    def hyper_dim(self):
        return self.prior.hyper_dim

    @property
    def pi_hyper(self):
        return self.prior.pi_hyper

    @property
    def pi_list(self):
        return self.prior.pi_list

    @property
    def ll_list(self):
        return [ll for ll,_ in self.logL.factors]

    @property
    def nsteps(self):
        return self.prior.nsteps

    def append(self, pi, ll=None):
        r""" Append a new transition distribution :math:`\pi({\bf Z}_{t_{k+1}}\vert\Theta, {\bf Z}_{t_{k}})` and the corresponding log-likelihood :math:`\log\mathcal{L}({\bf y}_{t_k} \vert \Theta, {\bf Z}_{t_k})` if any.

        Args:
          pi (:class:`ConditionalDistribution`): transition distribution
            :math:`\pi({\bf Z}_{t_{k+1}}\vert\Theta, {\bf Z}_{t_{k}})`
          ll (:class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
            log-likelihood
            :math:`\log\mathcal{L}({\bf y}_{t_k} \vert \Theta, {\bf Z}_{t_k})`.
            Missing data are represented by ``None``.
        """
        self.prior.append( pi )
        # Prepare likelihood factor and check consistency
        sdim = self.state_dim
        hdim = self.hyper_dim
        if ll is not None and ll.dim != sdim + hdim:
            raise ValueError("Dimension of log-likelihood is not consistent")
        self.logL.append( (ll, list(range(self.dim,self.dim+sdim)) + list(range(hdim))) )
        # Append
        self.dim = self.prior.dim

    def get_MarkovComponent(self, i, n=1, state_map=None, hyper_map=None):
        r""" Extract the (:math:`n\geq 1` steps) :math:`i`-th Markov component from the distribution

        If :math:`i=0` the Markov component is given by

        .. math::

           \pi^{0:n}(\Theta, {\bf Z}_{t_0}, \ldots, {\bf Z}_{t_n}) :=
             \left( \prod_{t \in \{t_0,\ldots,t_n\} \cap B}
                \mathcal{L}({\bf y}_t \vert \Theta, {\bf Z}_t) \right)
             \left( \prod_{i=1}^n \pi({\bf Z}_{t_i}\vert \Theta, {\bf Z}_{t_{i-1}}) \right)
             \pi({\bf Z}_{t_0}\vert\Theta) \pi(\Theta) \;.

        If :math:`i>0` then the Markov component is

        .. math::

           \pi^{i:i+n}\left(\Theta, {\bf Z}_{t_i}, \ldots, {\bf Z}_{t_{i+n}}\right) :=
             \eta(\Theta, {\bf Z}_{t_i}) 
             \left( \prod_{t \in \left\{t_{i+1},\ldots,t_{i+n}\right\} \cap B}
                \mathcal{L}\left({\bf y}_t \vert
                  \mathfrak{T}_{i-1}^{\Theta}(\Theta), {\bf Z}_t\right) \right)
             \left( \prod_{k=i+1}^{i+n-1}
                \pi\left({\bf Z}_{t_k+1}\vert {\bf Z}_{t_{k}},
                   \mathfrak{T}_{i-1}^{\Theta}(\Theta) \right) \right)
             \pi\left({\bf Z}_{t_{i+1}} \vert
                \mathfrak{M}_{i-1}^{1}(\Theta, {\bf Z}_{t_i}),
                \mathfrak{T}_{i-1}^{\Theta}(\Theta)
             \right) \;,

        where :math:`\mathfrak{T}_{i-1}^{\Theta}` and
        :math:`\mathfrak{M}_{i-1}^{1}` are the hyper-parameter and forecast
        components of the map computed at step :math:`i-1`, using the
        sequential algorithm described in :cite:`Spantini2017`.

        Args:
          i (int): index :math:`i` of the Markov component
          n (int): number of steps :math:`n`
          state_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
            forecast map :math:`\mathfrak{M}_{i-1}^{1}` from step :math:`i-1`.
          hyper_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
            hyper-parameter map :math:`\mathfrak{T}_{i-1}^{\Theta}`
            from step :math:`i-1`.

        Returns:
          (:class:`Distribution<TransportMaps.Distributions.Distribution>`) --
            Markov component :math:`\pi^{i:i+n}`.
        """
        if i == 0:
            out = MarkovComponentDistribution(
                i, self.pi_list[:n+1], self.ll_list[:n+1],
                self.state_dim, self.hyper_dim,
                self.pi_hyper)
        elif i > 0:
            out = MarkovComponentDistribution(
                i, self.pi_list[i+1:i+n+1], self.ll_list[i+1:i+n+1],
                self.state_dim, self.hyper_dim,
                self.pi_hyper,
                state_map, hyper_map)
        else:
            raise AttributeError("Index must be i >= 0")
        return out

    def trim(self, nsteps):
        r""" Trim the Markov chain to the first ``nsteps``

        Args:
          (int) nsteps: number of steps in the Markov chain of the returned
             distribution
        Returns:
          (:class:`SequentialHiddenMarkovChainDistribution`) -- trimmed distribution
        """
        return SequentialHiddenMarkovChainDistribution(
            self.pi_list[:nsteps],
            self.ll_list[:nsteps],
            self.pi_hyper)
        
class MarkovComponentDistribution(Distribution):
    r""" :math:`i`-th Markov component of a :class:`SequentialHiddenMarkovChainDistribution`

    If :math:`i=0` the Markov component is given by

    .. math::

       \pi^{0:n}(\Theta, {\bf Z}_{t_0}, \ldots, {\bf Z}_{t_n}) :=
         \left( \prod_{t \in \{t_0,\ldots,t_n\} \cap B}
            \mathcal{L}({\bf y}_t \vert \Theta, {\bf Z}_t) \right)
         \left( \prod_{i=1}^n \pi({\bf Z}_{t_i}\vert \Theta, {\bf Z}_{t_{i-1}}) \right)
         \pi({\bf Z}_{t_0}\vert\Theta) \pi(\Theta) \;.

    If :math:`i>0` then the Markov component is

    .. math::

       \pi^{i:i+n}\left(\Theta, {\bf Z}_{t_i}, \ldots, {\bf Z}_{t_{i+n}}\right) :=
         \eta(\Theta, {\bf Z}_{t_i}) 
         \left( \prod_{t \in \left\{t_{i+1},\ldots,t_{i+n}\right\} \cap B}
            \mathcal{L}\left({\bf y}_t \vert
              \mathfrak{T}_{i-1}^{\Theta}(\Theta), {\bf Z}_t\right) \right)
         \left( \prod_{k=i+1}^{i+n-1}
            \pi\left({\bf Z}_{t_k+1}\vert {\bf Z}_{t_{k}},
               \mathfrak{T}_{i-1}^{\Theta}(\Theta) \right) \right)
         \pi\left({\bf Z}_{t_{i+1}} \vert
            \mathfrak{M}_{i-1}^{1}(\Theta, {\bf Z}_{t_i}),
            \mathfrak{T}_{i-1}^{\Theta}(\Theta)
         \right) \;,

    where :math:`\mathfrak{T}_{i-1}^{\Theta}` and
    :math:`\mathfrak{M}_{i-1}^{1}` are the hyper-parameter and forecast
    components of the map computed at step :math:`i-1`, using the
    sequential algorithm described in :cite:`Spantini2017`.

    Args:
      idx0 (int): index :math:`i` of the Markov component
      pi_list (:class:`list` of :class:`Distribution`): list of :math:`n`
        transition densities
      ll_list (:class:`list` of :class:`LogLikelihood<TransportMaps.Functionals.LogLikelihood>`):
        list of :math:`n` log-likelihoods (``None`` for missing data)
        :math:`\{\log\mathcal{L}({\bf y}_t \vert \Theta, {\bf Z}_t)\}_{t\in B}`
      state_dim (int): dimension of the state-space
      hyper_dim (int): dimension of the parameter-space
      pi_hyper (:class:`Distribution`): prior on hyper-parameters :math:`h(\Theta)`
      state_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        forecast map :math:`\mathfrak{M}_{i-1}^{1}` from step :math:`i-1`.
      hyper_map (:class:`TransportMap<TransportMaps.Maps.TransportMap>`):
        hyper-parameter map :math:`\mathfrak{T}_{i-1}^{\Theta}`
        from step :math:`i-1`.
    """
    def __init__(self, idx0, pi_list, ll_list,
                 state_dim, hyper_dim,
                 pi_hyper=None,
                 state_map=None, hyper_map=None):
        # Figure out dimension and check correct arguments are provided
        self.idx0 = idx0
        self.state_dim = state_dim
        self.hyper_dim = hyper_dim
        if idx0 == 0:
            if pi_hyper is not None and pi_hyper.dim != hyper_dim:
                raise AttributeError(
                    "Dimension of prior of hyper-paramenter not consistent")
            self.pi_hyper = pi_hyper
        elif idx0 > 0:
            if state_map is None:
                raise AttributeError(
                    "The state_map parameter must be provided for idx0>0")
            if state_map.dim_in != hyper_dim + state_dim:
                raise AttributeError(
                    "Dimension of state_map not consistent")
            if self.hyper_dim > 0 and hyper_map is None:
                raise AttributeError(
                    "The hyper_map parameter must be provided for hyper_dim>0")
            if self.hyper_dim > 0 and hyper_map.dim != hyper_dim:
                raise AttributeError(
                    "Dimension of hyper_map not consistent")
            self.state_map = state_map
            self.hyper_map = hyper_map
            self.pi_hyper = StandardNormalDistribution(hyper_dim + state_dim)
        else:
            raise AttributeError("Onli idx0>=0 admitted")
        if len(pi_list) != len(ll_list):
            raise AttributeError("Length of list of transition must be the same of " + \
                                 "the list of log-likelihoods")
        if idx0 == 0 and len(pi_list) < 2:
            raise AttributeError("The 0-th Markov component must be composed " + \
                                 "by at least two transition densities")
        if idx0 > 0 and len(pi_list) < 1:
            raise AttributeError("The (i>0)-th Markov component must be composed " + \
                                 "by at least one transition distribution")
        if idx0 == 0: # We can get away with a SequentialInferenceDistribution
            self.seqinfdens = SequentialHiddenMarkovChainDistribution(
                pi_list, ll_list, pi_hyper)
            dim = self.seqinfdens.dim
        else:
            dim = hyper_dim + state_dim
            for i, (pi, ll) in enumerate(zip(pi_list,ll_list)):
                if pi.dim != state_dim or pi.dim_y != hyper_dim + state_dim:
                    raise AttributeError(
                        "The dimension of the %d transition is not consistent" % i)
                if ll is not None and ll.dim != hyper_dim + state_dim:
                    raise AttributeError(
                        "The dimension of the %d log-likelihood is not consistent" % i)
                dim += state_dim
        self.pi_list = pi_list
        self.ll_list = ll_list
        super(MarkovComponentDistribution, self).__init__(dim)

    def log_pdf(self, x, *args, **kwargs):
        if self.idx0 == 0:
            out = self.seqinfdens.log_pdf(x, *args, **kwargs)
        else:
            xtr = x.copy() # Transformed
            hdim = self.hyper_dim
            sdim = self.state_dim
            # Apply hyper and component maps to obtain transformed inputs
            if hdim > 0:
                xtr[:,:hdim] = self.hyper_map.evaluate(x[:,:hdim], *args, **kwargs)
            xtr[:,hdim:hdim+sdim] = self.state_map.evaluate(
                x[:,:hdim+sdim], *args, **kwargs)
            # Evaluate pi_hyper
            out = self.pi_hyper.log_pdf(x[:,:hdim+sdim], *args, **kwargs)
            # Evaluate transitions and likelihoods
            for i, (pi, ll) in enumerate(zip(self.pi_list, self.ll_list)):
                s1 = hdim + i*sdim
                s2 = s1 + sdim
                s3 = s2 + sdim
                xin = xtr[:,s2:s3]
                yin = np.hstack( (xtr[:,s1:s2], xtr[:,:hdim]) )
                out += pi.log_pdf(xin, yin, *args, **kwargs)
                if ll is not None:
                    xin = np.hstack( (xtr[:,s2:s3], xtr[:,:hdim]) )
                    out += ll.evaluate(xin, *args, **kwargs)
        return out

    def grad_x_log_pdf(self, x, *args, **kwargs):
        if self.idx0 == 0:
            out = self.seqinfdens.grad_x_log_pdf(x, *args, **kwargs)
        else:
            xtr = x.copy() # Transformed input
            hdim = self.hyper_dim
            sdim = self.state_dim
            # Apply hyper and component maps to obtain transformed inputs and
            # compute gradients of the maps
            if hdim > 0:
                xtr[:,:hdim] = self.hyper_map.evaluate(x[:,:hdim], *args, **kwargs)
                gx_hyper = self.hyper_map.grad_x(x[:,:hdim], *args, **kwargs)
            xtr[:,hdim:hdim+sdim] = self.state_map.evaluate(
                x[:,:hdim+sdim], *args, **kwargs)
            gx_comp = self.state_map.grad_x(x[:,:hdim+sdim], *args, **kwargs)
            # Evaluate
            out = np.zeros((x.shape[0], self.dim))
            # Evaluate pi_hyper
            out[:,:hdim+sdim] += self.pi_hyper.grad_x_log_pdf(
                x[:,:hdim+sdim], *args, **kwargs)
            # Evaluate transitions and likelihoods
            perm_pi = list(range(2*sdim,2*sdim+hdim)) + \
                      list(range(sdim,2*sdim)) + \
                      list(range(0, sdim))
            perm_ll = list(range(sdim, sdim+hdim)) + \
                      list(range(0, sdim))
            for i, (pi, ll) in enumerate(zip(self.pi_list, self.ll_list)):
                s1 = hdim + i*sdim
                s2 = s1 + sdim
                s3 = s2 + sdim
                xin = xtr[:,s2:s3]
                yin = np.hstack( (xtr[:,s1:s2], xtr[:,:hdim]) )
                # Transition
                gxlpdf = pi.grad_x_log_pdf(xin, yin, *args, **kwargs)[:,perm_pi]
                gxlpdfinner = np.zeros(gxlpdf.shape)
                gxlpdfinner[:,hdim+sdim:] = gxlpdf[:,hdim+sdim:]
                if hdim > 0: # Apply gradient hyper map
                    gxlpdfinner[:,:hdim] += np.einsum(
                        '...i,...ij->...j', gxlpdf[:,:hdim], gx_hyper)
                if i == 0: # Apply gradient component map
                    gxlpdfinner[:,:hdim+sdim] += np.einsum(
                        '...i,...ij->...j', gxlpdf[:,hdim:hdim+sdim], gx_comp)
                else:
                    gxlpdfinner[:,hdim:hdim+sdim] = gxlpdf[:,hdim:hdim+sdim]
                out[:,:hdim] += gxlpdfinner[:,:hdim]
                out[:,s1:s3] += gxlpdfinner[:,hdim:]
                # Likelihood
                if ll is not None:
                    xin = np.hstack( (xtr[:,s2:s3], xtr[:,:hdim]) )
                    gxll = ll.grad_x(xin, *args, **kwargs)[:,perm_ll]
                    if hdim > 0:
                        gxll[:,:hdim] = np.einsum(
                            '...i,...ij->...j', gxll[:,:hdim], gx_hyper)
                    out[:,:hdim] += gxll[:,:hdim]
                    out[:,s2:s3] += gxll[:,hdim:]
        return out

    def hess_x_log_pdf(self, x, *args, **kwargs):
        if self.idx0 == 0:
            out = self.seqinfdens.hess_x_log_pdf(x, *args, **kwargs)
        else:
            xtr = x.copy() # Transformed input
            hdim = self.hyper_dim
            sdim = self.state_dim
            # Apply hyper and component maps to obtain transformed inputs and
            # compute gradients and Hessians of the maps
            xtr[:,hdim:hdim+sdim] = self.state_map.evaluate(
                x[:,:hdim+sdim], *args, **kwargs)
            gx_comp = self.state_map.grad_x(x[:,:hdim+sdim], *args, **kwargs)
            hx_comp = self.state_map.hess_x(x[:,:hdim+sdim], *args, **kwargs)
            if hdim > 0:
                xtr[:,:hdim] = self.hyper_map.evaluate(x[:,:hdim], *args, **kwargs)
                gx_hyper = self.hyper_map.grad_x(x[:,:hdim], *args, **kwargs)
                hx_hyper = self.hyper_map.hess_x(x[:,:hdim], *args, **kwargs)
            # Evaluate
            m = x.shape[0]
            out = np.zeros((x.shape[0], self.dim, self.dim))
            # Evaluate pi_hyper
            out[:,:hdim+sdim,:hdim+sdim] += self.pi_hyper.hess_x_log_pdf(
                x[:,:hdim+sdim], *args, **kwargs)
            # Evaluate transitions and likelihoods
            perm_pi = list(range(2*sdim,2*sdim+hdim)) + \
                      list(range(sdim,2*sdim)) + \
                      list(range(0, sdim))
            perm_ll = list(range(sdim, sdim+hdim)) + \
                      list(range(0, sdim))
            for i, (pi, ll) in enumerate(zip(self.pi_list, self.ll_list)):
                s1 = hdim + i*sdim
                s2 = s1 + sdim
                s3 = s2 + sdim
                xin = xtr[:,s2:s3]
                yin = np.hstack( (xtr[:,s1:s2], xtr[:,:hdim]) )
                #############
                # Transition
                c1 = hdim
                c2 = c1 + sdim
                hxlpdf = pi.hess_x_log_pdf(
                    xin, yin, *args, **kwargs)[np.ix_(range(m),perm_pi,perm_pi)]
                gxlpdf = pi.grad_x_log_pdf(xin, yin, *args, **kwargs)[:,perm_pi]
                hx = np.zeros(hxlpdf.shape)

                # H_th (logpi(T(th), M(th,y), x))
                if hdim > 0: # Apply hyper map
                    # <H_th logpi, (d_th T)(d_th T)^T>
                    tmp = np.einsum('...ij,...ik->...jk', hxlpdf[:,:c1,:c1], gx_hyper)
                    tmp = np.einsum('...ij,...ik->...jk', tmp, gx_hyper)
                    hx[:,:c1,:c1] += tmp
                    if i == 0:
                        # <H_{th,y} logpi, (d_th T)(d_th M)^T>
                        tmp = np.einsum('...ij,...ik->...jk', hxlpdf[:,:c1,c1:c2], gx_hyper)
                        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_comp[:,:,:c1])
                        hx[:,:c1,:c1] += tmp + tmp.transpose((0,2,1)) # Symmetry
                        # <H_{y} logpi, (d_th M)(d_th M)^T>
                        tmp = np.einsum('...ij,...ik->...jk',
                                        hxlpdf[:,c1:c2,c1:c2], gx_comp[:,:,:c1])
                        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_comp[:,:,:c1])
                        hx[:,:c1,:c1] += tmp
                    # <d_th logpi, H_th T>
                    tmp = np.einsum('...i,...ijk->...jk', gxlpdf[:,:c1], hx_hyper)
                    hx[:,:c1,:c1] += tmp
                    if i == 0:
                        # <d_y logpi, H_th M>
                        tmp = np.einsum('...i,...ijk->...jk',
                                        gxlpdf[:,c1:c2], hx_comp[:,:,:c1,:c1])
                        hx[:,:c1,:c1] += tmp

                # H_y (logpi(T(th), M(th,y), x))
                if i == 0:
                    # <H_y logpi, (d_y M)(d_y M)^T>
                    tmp = np.einsum('...ij,...ik->...jk',
                                    hxlpdf[:,c1:c2,c1:c2], gx_comp[:,:,c1:])
                    tmp = np.einsum('...ij,...ik->...jk', tmp, gx_comp[:,:,c1:])
                    hx[:,c1:c2,c1:c2] += tmp
                    # <d_y logpi, H_y M>
                    tmp = np.einsum('...i,...ijk->...jk',
                                    gxlpdf[:,c1:c2], hx_comp[:,:,c1:,c1:])
                    hx[:,c1:c2,c1:c2] += tmp
                else:
                    hx[:,c1:c2,c1:c2] += hxlpdf[:,c1:c2,c1:c2]

                # H_x (logpi(T(th), M(th,y), x))
                hx[:,c2:,c2:] += hxlpdf[:,c2:,c2:] # Identity

                # H_{th,y} (logpi(T(th), M(th,y), x))
                if hdim > 0:
                    if i == 0:
                        # <H_{th,y} logpi, (d_th T)(d_y M)^T>
                        tmp = np.einsum('...ij,...ik->...jk', hxlpdf[:,:c1,c1:c2], gx_hyper)
                        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_comp[:,:,c1:])
                        hx[:,:c1,c1:c2] += tmp
                        # <H_y logpi, (d_th M)(d_y M)^T>
                        tmp = np.einsum('...ij,...ik->...jk',
                                        hxlpdf[:,c1:c2,c1:c2], gx_comp[:,:,:c1])
                        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_comp[:,:,c1:])
                        hx[:,:c1,c1:c2] += tmp
                        # <d_y logpi, H_{th,y} M>
                        tmp = np.einsum('...i,...ijk->...jk',
                                        gxlpdf[:,c1:c2], hx_comp[:,:,:c1,c1:])
                        hx[:,:c1,c1:c2] += tmp
                    else:
                        # <H_{th,y} logpi, (d_th T) I^T>
                        tmp = np.einsum('...ij,...ik->...kj', hxlpdf[:,:c1,c1:c2], gx_hyper)
                        hx[:,:c1,c1:c2] += tmp

                # H_{y,th} (logpi(T(th), M(th,y), x)) -- symmetry
                hx[:,c1:c2,:c1] += hx[:,:c1,c1:c2].transpose((0,2,1))

                # H_{th,x} (logpi(T(th), M(th,y), x))
                if hdim > 0:
                    # <H_{th,x} logpi, (d_th T) I^T>
                    tmp = np.einsum('...ij,...ik->...kj', hxlpdf[:,:c1,c2:], gx_hyper)
                    hx[:,:c1,c2:] += tmp
                    if i == 0:
                        tmp = np.einsum('...ij,...ik->...kj',
                                        hxlpdf[:,c1:c2,c2:], gx_comp[:,:,:c1])
                        hx[:,:c1,c2:] += tmp

                # H_{x,th} (logpi(T(th), M(th,y), x)) -- symmetry
                hx[:,c2:,:c1] += hx[:,:c1,c2:].transpose((0,2,1))

                # H_{y,x} (logpi(T(th), M(th,y), x))
                if i == 0:
                    # <H_{y,x} logpi, (d_y M) I^T>
                    tmp = np.einsum('...ij,...ik->...kj',
                                    hxlpdf[:,c1:c2,c2:], gx_comp[:,:,c1:])
                    hx[:,c1:c2,c2:] += tmp
                else:
                    hx[:,c1:c2,c2:] += hxlpdf[:,c1:c2,c2:]

                # H_{x,y} (logpi(T(th), M(th,y), x)) -- symmetry
                hx[:,c2:,c1:c2] = hx[:,c1:c2,c2:].transpose((0,2,1))

                # Update out
                out[:,:hdim,:hdim] += hx[:,:hdim,:hdim]
                out[:,:hdim,s1:s3] += hx[:,:hdim,hdim:]
                out[:,s1:s3,:hdim] += hx[:,hdim:,:hdim]
                out[:,s1:s3,s1:s3] += hx[:,hdim:,hdim:]
                
                #############
                # Likelihood
                if ll is not None:
                    xin = np.hstack( (xtr[:,s2:s3], xtr[:,:hdim]) )
                    hxll = ll.hess_x(
                        xin, *args, **kwargs)[np.ix_(range(m),perm_ll,perm_ll)]
                    gxll = ll.grad_x(xin, *args, **kwargs)[:,perm_ll]
                    hx = np.zeros(hxll.shape)

                    # H_th (logL(T(th),x))
                    if hdim > 0:
                        # <H_th logL, (d_th T)(d_th T)^T>
                        tmp = np.einsum('...ij,...ik->...jk', hxll[:,:hdim,:hdim], gx_hyper)
                        tmp = np.einsum('...ij,...ik->...jk', tmp, gx_hyper)
                        hx[:,:hdim,:hdim] += tmp
                        # <d_th logL, H_th T>
                        tmp = np.einsum('...i,...ijk->...jk', gxll[:,:hdim], hx_hyper)
                        hx[:,:hdim,:hdim] += tmp

                    # H_x (logL(T(th),x))
                    hx[:,hdim:,hdim:] += hxll[:,hdim:,hdim:]

                    # H_{th,x} (logL(T(th),x))
                    if hdim > 0:
                        # <H_{th,x} logL, (d_th T) I^T>
                        tmp = np.einsum('...ij,...ik->...kj', hxll[:,:hdim,hdim:], gx_hyper)
                        hx[:,:hdim,hdim:] += tmp

                    # H_{th,x} (logL(T(th),x)) -- symmetry
                    hx[:,hdim:,:hdim] += hx[:,:hdim,hdim:].transpose((0,2,1))

                    # Update out
                    out[:,:hdim,:hdim] += hx[:,:hdim,:hdim]
                    out[:,:hdim,s2:s3] += hx[:,:hdim,hdim:]
                    out[:,s2:s3,:hdim] += hx[:,hdim:,:hdim]
                    out[:,s2:s3,s2:s3] += hx[:,hdim:,hdim:]
        return out


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import collections
import numpy as np
import scipy.linalg as scila

from .SequentialInferenceBase import *

from TransportMaps.Distributions import GaussianDistribution
from TransportMaps.Likelihoods import AdditiveLinearGaussianLogLikelihood
from TransportMaps.Maps import LinearTransportMap, PermutationTransportMap, \
    ListCompositeTransportMap
from TransportMaps.Maps.Decomposable import LiftedTransportMap

__all__ = ['LinearFilter',
           'LinearSmoother']

class LinearFilter(Filter):
    r""" Perform the on-line filtering of a sequential linear Gaussian Hidden Markov chain.

    Aka: Kalman filter.

    If the linear state-space model is parametric, i.e.

    .. math::

       {\bf Z}_{k+1} = {\bf c}_k(\theta) + {\bf F}_k(\theta){\bf Z}_k + {\bf w}_k(\theta) \\
       {\bf Y}_{k} = {\bf H}_k(\theta){\bf Z}_k + {\bf v}_k(\theta)

    then one can optionally compute the gradient
    (with respect to the parameters) of the filter.

    Args:
      ders (int): ``0`` no gradient is computed, ``1`` compute gradient
      pi_hyper (:class:`Distribution<TransportMaps.Distributions.Distribution>`):
        prior distribution on the hyper-parameters :math:`\pi(\Theta)`
    
    .. todo:: Square-root filter
    """
    def __init__(self, ders=0, pi_hyper=None):
        super(LinearFilter, self).__init__(pi_hyper)
        self._marg_log_lklhood = 0.
        self._filt_mean_list = []
        self._filt_cov_list = []
        self._ders = ders
        if ders > 0:
            self._filt_grad_mean_list = []
            self._filt_grad_cov_list = []
            self._grad_marg_log_lklhood = 0.

    @property
    def marginal_log_likelihood(self):
        r""" Returns the marginal log-likelihood :math:`\log\pi\left({\bf Y}_{\Xi\leq k}\right)`

        Returns:
          (:class:`float`) -- current marginal likelihood
        """
        return self._marg_log_lklhood
        
    @property
    def filtering_mean_list(self):
        r""" Returns the means of all the filtering distributions

        Returns:
          (:class:`list` of :class:`float`) -- means of
            :math:`\pi\left({\bf Z}_k\middle\vert{\bf Y}_{\Xi\leq k}\right)`
            for :math:`k\in \Lambda=0,\ldots,n`.
        """
        return self._filt_mean_list

    @property
    def filtering_covariance_list(self):
        r""" Returns the covariances of all the filtering distributions

        Returns:
          (:class:`list` of :class:`ndarray<numpy.ndarray>`) -- covariances of
            :math:`\pi\left({\bf Z}_k\middle\vert{\bf Y}_{\Xi\leq k}\right)`
            for :math:`k\in \Lambda=0,\ldots,n`.
        """
        return self._filt_cov_list

    @property
    def grad_marginal_log_likelihood(self):
        r""" Returns the gradient of the marginal log-likelihood :math:`\nabla_\theta\log\pi\left({\bf Y}_{\Xi\leq k}\right)`

        Returns:
          (:class:`float`) -- current marginal likelihood
        """
        return self._grad_marg_log_lklhood
        
    @property
    def filtering_grad_mean_list(self):
        r""" Returns the gradient of the means of all the filtering distributions

        Returns:
          (:class:`list` of :class:`float`) -- gradient of the means of
            :math:`\pi\left({\bf Z}_k\middle\vert{\bf Y}_{\Xi\leq k}\right)`
            for :math:`k\in \Lambda=0,\ldots,n`.
        """
        return self._filt_grad_mean_list

    @property
    def filtering_grad_covariance_list(self):
        r""" Returns the gradient of the covariances of all the filtering distributions

        Returns:
          (:class:`list` of :class:`ndarray<numpy.ndarray>`) --
            gradient of the covariances of
            :math:`\pi\left({\bf Z}_k\middle\vert{\bf Y}_{\Xi\leq k}\right)`
            for :math:`k\in \Lambda=0,\ldots,n`.
        """
        return self._filt_grad_cov_list


    def _assimilation_step(self):
        r""" Assimilate one piece of Gaussian data :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}\right) \right)`.
        """
        pi = self.pi.pi_list[-1]
        ll = self.pi.ll_list[-1]

        # PRECIDTION STEP
        if self.nsteps == 0:
            # Load initial conditions x_{0|-1} and P_{0|-1}
            pred_x = pi.mu
            pred_P = pi.sigma
        else:
            # Load former state and covariance
            x = self._filt_mean_list[-1]
            P = self._filt_cov_list[-1]
            # Load transition system
            c = pi.T.c
            F = pi.T.T
            b = pi.pi.mu
            Q = pi.pi.sigma
            # Predict
            pred_x = c + F.dot(x) + b
            pred_P = F.dot(P.dot(F.T)) + Q
            
        # Gradient recursion
        if self._ders > 0:
            # Prediction step
            if self.nsteps == 0:
                pred_ga_x = pi.grad_a_mu
                pred_ga_P = pi.grad_a_sigma
            else:
                # Load former gradient state and gradient covariance
                ga_x = self._filt_grad_mean_list[-1]
                ga_P = self._filt_grad_cov_list[-1]
                # Load transition derivatives
                ga_c = pi.T.grad_a_c
                ga_F = pi.T.grad_a_T
                ga_b = pi.pi.grad_a_mu
                ga_Q = pi.pi.grad_a_sigma
                # Predict
                pred_ga_x = ga_c + np.einsum('ijk,j->ik', ga_F, x) + \
                            np.einsum('ij,jk->ik', F, ga_x) + ga_b
                # gF P F^T + F gP F^T + F P gF^T
                pred_ga_P = np.einsum('ijk,jl,ml->imk', ga_F, P, F) + \
                            np.einsum('ij,jkl,mk->iml', F, ga_P, F) + \
                            np.einsum('ij,jk,lkm->ilm', F, P, ga_F) + \
                            ga_Q

        # State and convariance after prediction
        x = pred_x
        P = pred_P
        if self._ders > 0:
            ga_x = pred_ga_x
            ga_P = pred_ga_P
            
        # Perform update if necessary
        up_x = x
        up_P = P
        if ll is not None:
            # Load observation system
            z = ll.y
            a = ll.T.c
            H = ll.T.T
            b = ll.pi.mu
            R = ll.pi.sigma
            # Compute update recursions
            y = z - a - b - np.dot(H, x) # Mesurament residual
            S = R + np.dot(H, np.dot(P, H.T)) # Innovation
            S12_factors = scila.cho_factor(S, True)
            S1y = scila.cho_solve(S12_factors, y)
            S1H = scila.cho_solve(S12_factors, H)
            KS = np.dot(P, H.T)
            Ky = np.dot(KS, S1y)
            KH = np.dot(KS, S1H)
            ImKH = np.eye(pi.dim) - KH
            # Upadtes (state, covariance, marginal likelihood)
            up_x = x + Ky
            up_P = np.dot(ImKH, P)
            self._marg_log_lklhood -= .5 * (np.dot(y, S1y) + \
                                            2 * np.sum(np.log(np.diag(S12_factors[0]))) + \
                                            y.shape[0] * np.log(2*np.pi))

        # Gradient recursion
        if ll is not None and self._ders > 0:
            # Load gradients of observation system
            ga_a = ll.T.grad_a_c
            ga_H = ll.T.grad_a_T
            ga_b = ll.pi.grad_a_mu
            ga_R = ll.pi.grad_a_sigma
            ncoeffs = ga_a.shape[1]
            # Compute update recursions
            ga_y = - (ga_a + ga_b + np.einsum('ijk,j->ik', ga_H, x) + \
                      np.einsum('ij,jk->ik', H, ga_x) )
            ga_S = ga_R + \
                   np.einsum('ijk,jl,ml->imk', ga_H, P, H) + \
                   np.einsum('ij,jkl,mk->iml', H, ga_P, H) + \
                   np.einsum('ij,jk,lkm->ilm', H, P, ga_H)
            S1gaS = np.zeros( ga_S.shape )
            for nc in range(ncoeffs):
                S1gaS[:,:,nc] = scila.cho_solve(S12_factors, ga_S[:,:,nc])
            ga_KS = np.einsum('ijk,lj->ilk', ga_P, H) + \
                    np.einsum('ij,kjl->ikl', P, ga_H) + \
                    np.einsum('ij,kj,klm->ilm', P, H, - S1gaS)
            S1gay = np.zeros( ga_y.shape )
            for nc in range(ncoeffs):
                S1gay[:,nc] = scila.cho_solve(S12_factors, ga_y[:,nc])
            S1gaH = np.zeros( ga_H.shape )
            for nc in range(ncoeffs):
                S1gaH[:,:,nc] = scila.cho_solve(S12_factors, ga_H[:,:,nc])
            Kgay = np.dot(KS, S1gay)
            gaKy = np.einsum('ijk,j->ik', ga_KS, S1y)
            gaKH = np.einsum('ijk,jl->ilk', ga_KS, S1H)
            KgaH = np.einsum('ij,jkl->ikl', KS, S1gaH)
            # Upadtes (state, covariance)
            up_ga_x = ga_x + gaKy + Kgay
            up_ga_P = - np.einsum('ijk,jl->ilk', gaKH + KgaH, P) \
                      + np.einsum('ij,jkl->ikl', ImKH, ga_P)
            # Update marginal likelihood
            tr_S1gaS = np.einsum('iik->k', S1gaS) # Trace for each parameter
            ga_mll = .5 * (np.dot(ga_y.T, S1y) + \
                           np.einsum('i,ijk,j->k', y, -S1gaS, S1y) + \
                           np.dot(y, S1gay) + \
                           tr_S1gaS)
            self._grad_marg_log_lklhood -= ga_mll
            
        # State and convariance after update
        x = up_x
        P = up_P
        if self._ders > 0:
            ga_x = up_ga_x
            ga_P = up_ga_P
            
        # Append new filtering map
        try:
            L = scila.cholesky(P, True)
        except scila.LinAlgError:
            U,S,V = scila.svd(P)
            L = U * np.sqrt(S)
        M = LinearTransportMap(x, L)
        self.F_list.append(M)

        # Update filtering mean and covariance
        self._filt_mean_list.append(x)
        self._filt_cov_list.append( np.dot(L, L.T) )
        if self._ders > 0:
            self._filt_grad_mean_list.append( ga_x )
            self._filt_grad_cov_list.append( ga_P )

class LinearSmoother(LinearFilter,Smoother):
    r""" Perform the on-line assimilation of a sequential linear Gaussian Hidden Markov chain.

    Args:
      lag (:class:`numpy.float`): lag to be used in the backward updates of
        smoothing means and covariances. The default value ``None`` indicates
        infinite lag.
    
    .. todo:: no hyper-parameter admitted right now.
    """
    def __init__(self, lag=None):
        super(LinearSmoother,self).__init__()
        self._lag = lag
        self._smooth_mean_list = []
        self._smooth_cov_list = []
        # Data structures necessary for fast smoothing updates
        self._CB_queue = collections.deque(maxlen=lag)
        self._CB2_queue = collections.deque(maxlen=lag)

    @property
    def lag(self):
        return self._lag

    @property
    def smoothing_mean_list(self):
        return self._smooth_mean_list

    @property
    def smoothing_covariance_list(self):
        return self._smooth_cov_list

    def _assimilation_step(self):
        r""" Assimilate one piece of Gaussian data :math:`\left( \pi\left({\bf Z}_{k+1} \middle\vert {\bf Z}_k \right), \log \mathcal{L}\left({\bf y}_{k+1}\middle\vert {\bf Z}_{k+1}\right) \right)`.
        """
        # Run the filter to get (c,L)
        LinearFilter._assimilation_step(self)
        c = self.F_list[-1].c
        C = self.F_list[-1].L
        a = None
        A = None
        B = None
        # Apply recursions
        if self.nsteps == 0: # At first step smoothing == filtering
            M = LinearTransportMap(c, C)
            L = LiftedTransportMap(-1, M, self.pi.dim, 0)
            self.L_list.append( L )
        else:
            cm1 = self.F_list[-2].c
            Cm1 = self.F_list[-2].L
            # Load transition system
            pi = self.pi.pi_list[-1]
            sdim = pi.dim
            F = pi.T.T
            Q12 = pi.pi.sampling_mat
            # Recursions
            Q12FCm1 = scila.solve_triangular(Q12, np.dot(F, Cm1), lower=True)
            Q1FCm1 = scila.solve_triangular(Q12, Q12FCm1, lower=True, trans='T')
            J = np.eye(pi.dim) + np.dot(Cm1.T, np.dot(F.T, Q1FCm1))
            J12 = scila.cholesky(J, lower=True)
            Cm1TFT = np.dot(Cm1.T, F.T)
            # A
            A = scila.solve_triangular(J12, np.eye(pi.dim), lower=True, trans='T')
            # B
            Q12C = scila.solve_triangular(Q12, C, lower=True)
            Q1C = scila.solve_triangular(Q12, Q12C, lower=True, trans='T')
            PC = - np.dot(Cm1TFT,Q1C)
            J12PC = scila.solve_triangular(J12, PC, lower=True)
            B = - scila.solve_triangular(J12, J12PC, lower=True, trans='T')
            # a
            Fcm1c = np.dot(F,cm1) - c
            Q12Fcm1c = scila.solve_triangular(Q12, Fcm1c, lower=True)
            Q1Fcm1c = scila.solve_triangular(Q12, Q12Fcm1c, lower=True, trans='T')
            PFcm1c = - np.dot(Cm1TFT, Q1Fcm1c)
            J12PFcm1c = scila.solve_triangular(J12, PFcm1c, lower=True)
            a = scila.solve_triangular(J12, J12PFcm1c, lower=True, trans='T')
            # Assemble smoothing map
            ac = np.hstack((a, c))
            ABC = np.zeros((2*sdim, 2*sdim)) # Matrix [[A,B],[0,C]]
            ABC[:sdim,:sdim] = A
            ABC[:sdim,sdim:] = B
            ABC[sdim:,sdim:] = C
            M = LinearTransportMap(ac, ABC)

            # Update all dimension of previous maps
            for L in self.L_list:
                L.dim = L.dim_in = L.dim_out = self.pi.dim
            # Append new map
            L = LiftedTransportMap(self.nsteps-1, M, self.pi.dim, 0)
            self.L_list.append(L)

        # Update smoothing means and covariances
        self._update_smoothing_mean_covariance_lists(
            self.nsteps, self.lag,
            self._smooth_mean_list, self._smooth_cov_list,
            self._CB_queue, self._CB2_queue,
            c, C, a, A, B)

    def _update_smoothing_mean_covariance_lists(
            self, nsteps, lag,
            smooth_mean_list, smooth_cov_list, CB_queue, CB2_queue,
            c, C, a=None, A=None, B=None):
        # (the lag is controlled by the length of CB_queue and CB2_queue)
        for i, (CB, CB2) in enumerate(zip(CB_queue, CB2_queue)):
            if lag is None:
                lag = nsteps
            idx = max(nsteps - lag, 0) + i
            # MEAN: Insert new term <CB,a>
            smooth_mean_list[idx] += np.dot(CB, a)
            # COV: Remove last <CB,CB.T> term introduced
            smooth_cov_list[idx] -= CB2
            # COV: Insert new term <CBA,CBA.T>
            CBA = np.dot(CB, A)
            CBA2 = np.dot(CBA, CBA.T)
            smooth_cov_list[idx] += CBA2
            # Update CB and CB2
            CBB = np.dot(CB, B)
            CB_queue[i] = CBB
            CBB2 = np.dot(CBB, CBB.T)
            CB2_queue[i] = CBB2
            # COV: Insert new term <CBB,CBB.T>
            smooth_cov_list[idx] += CBB2
        # Append new mean
        smooth_mean_list.append( c.copy() )
        # Append new covariance
        CB_queue.append( C )
        C2 = np.dot(C, C.T)
        CB2_queue.append( C2 )
        smooth_cov_list.append( C2.copy() )
            
    def offline_smoothing_mean_covariance_lists(self, lag=None):
        r""" Compute the mean and covariances with a fixed lag for a pre-assimilated density
        """
        sdim = self.pi.state_dim
        smooth_mean_list = []
        smooth_cov_list = []
        # Data structures necessary for fast smoothing updates
        CB_queue = collections.deque(maxlen=lag)
        CB2_queue = collections.deque(maxlen=lag)
        # Iterate over the available maps
        for nsteps, L in enumerate(self.L_list):
            M = L.tm
            if nsteps > 0:
                c = M.c[sdim:]
                C = M.L[sdim:,sdim:]
                a = M.c[:sdim]
                A = M.L[:sdim,:sdim]
                B = M.L[:sdim,sdim:]
            else:
                c = M.c[:sdim]
                C = M.L[:sdim,:sdim]
                a = None
                A = None
                B = None
            self._update_smoothing_mean_covariance_lists(
                nsteps, lag,
                smooth_mean_list, smooth_cov_list,
                CB_queue, CB2_queue,
                c, C, a, A, B)
        return (smooth_mean_list, smooth_cov_list)


#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import numpy.linalg as npla
import scipy.optimize as sciopt
import scipy.linalg as scila
from  TransportMaps.Distributions import *
import sys
import scipy.sparse as scisp
import copy as cp


from TransportMaps.Maps.TransportMapBase import *

__all__ = ['LinearTransportMap']

class LinearTransportMap(TransportMap):
    r""" Linear map :math:`T({\bf x})={\bf c} + {\bf L}{\bf x}`

    Args:
      constantTerm (:class:`ndarray<numpy.ndarray>` [:math:`d`]): term :math:`{\bf c}`
      linearTerm (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]): term :math:`{\bf L}`
      log_det (float): :math:`\log\det{\bf L}`, only required if
        :math:`{\bf L}` is in the sparse format
        :module:`scipy.sparse<scipy.sparse>`.
    """
    def __init__(self, constantTerm, linearTerm, log_det = None):
        if len(constantTerm) != linearTerm.shape[0]:
            raise ValueError("Inconsistent dimensions")
        if linearTerm.shape[0] != linearTerm.shape[1]:
            raise ValueError("Inconsistent dimensions")

        if scisp.issparse(linearTerm):
            if log_det is None:
                raise ValueError("We did not implement the log-determinant of sparse matrices yet")
        else:
            #Make sure it is monotone increasing and invertible
            p,l,u  = scila.lu(linearTerm)
            diagu = np.nonzero( abs(np.diag(u))<1e-12 )
            if len(diagu[0])>0:
                raise ValueError("The map might not be increasing")
            sign_log_det , log_det =  np.linalg.slogdet(linearTerm)


        # if sign_log_det<0:
        #   raise ValueError("The map should be monotone increasing")
        self.dim = self.dim_in = self.dim_out = len(constantTerm)
        self.constantTerm = cp.deepcopy(constantTerm)
        self.linearTerm = cp.deepcopy(linearTerm) #Make sure the copy is valid even when the matrices are sparse
        self.log_det_linearTerm = log_det

    @property
    def c(self):
        r""" The constant term :math:`{\bf c}`
        """
        return self.constantTerm

    @c.setter
    def c(self, value):
        if len(value) != self.dim_out:
            raise ValueError("Inconsistent dimensions")
        self.constantTerm = value
        
    @property
    def L(self):
        r""" The linear term :math:`{\bf L}`
        """
        return self.linearTerm

    @L.setter
    def L(self, value):
        if self.dim != value.shape[0]:
            raise ValueError("Inconsistent dimensions")
        if value.shape[0] != value.shape[1]:
            raise ValueError("Inconsistent dimensions")
        #Make sure it is monotone increasing and invertible
        p,l,u  = scila.lu(value)
        diagu = np.nonzero( abs(np.diag(u))<1e-12 )
        if len(diagu[0])>0:
          raise ValueError("The map might not be increasing")
        sign_log_det , log_det =  np.linalg.slogdet(value)
        if sign_log_det<0:
          raise ValueError("The map should be monotone increasing")
        self.linearTerm = value
        self.log_det_linearTerm = log_det
        
    @staticmethod
    def build_from_Gaussian(pi , typeMap = "Sym"):
        r""" Build a linear transport map from a
        standard normal to a Gaussian distribution pi

        Args:
           pi (:class:`GaussianDistribution`):
              constant term of the linear map
        Raises:
           ValueError: if the shape of linear and constant term are inconsistent.
        """
        if not isinstance(pi, GaussianDistribution):
          raise ValueError("The input distribution should be a Gaussian")
        if typeMap == "Sym":
          try:
              U, s, V = np.linalg.svd(pi.sigma)
              s_sr = np.sqrt(s)
              linearTerm_sr = np.dot( U , np.diag( np.sqrt(s_sr) ) )
              linearTerm = np.dot( linearTerm_sr, linearTerm_sr.transpose() )
          except AttributeError:
              U, s, V = np.linalg.svd(pi.inv_sigma)
              s_inv_sr = np.sqrt(1./s)
              linearTerm_sr = np.dot( U , np.diag( np.sqrt(s_inv_sr) ) )
              linearTerm = np.dot( linearTerm_sr, linearTerm_sr.transpose() )
        elif typeMap == "Tri":
          try:
              linearTerm = np.linalg.cholesky(pi.sigma)
          except AttributeError:
              sigma = np.linalg.inv(pi.inv_sigma)
              linearTerm = np.linalg.cholesky(sigma)
        else:
           raise ValueError("Type of the map not supported yet")
        constantTerm = pi.mu
        return LinearTransportMap(constantTerm, linearTerm)

    @property
    def coeffs(self):
        r""" Returns the constant and linear term of the linear map.

        Returns:
           (:class:`ndarray<numpy.ndarray>`, :class:`ndarray<numpy.ndarray>`) --
           tuple (constant term, linear term)
        """
        return self.constantTerm, self.linearTerm

    @coeffs.setter
    def set_coeffs(self, constantTerm, linearTerm):
        r""" Set the constant and linear term of the linear map.

        Args:
           constantTerm (:class:`ndarray<numpy.ndarray>`):
              constant term of the linear map
           linearTerm (:class:`ndarray<numpy.ndarray>`):
              linear term of the linear map
        Raises:
           ValueError: if the shape of linear and constant term are inconsistent.
        """
        if len(constantTerm) != linearTerm.shape[0]:
            raise ValueError("Inconsistent dimensions")
        if linearTerm.shape[0] != linearTerm.shape[1]:
            raise ValueError("Inconsistent dimensions")
        #Make sure it is monotone increasing and invertible
        p,l,u  = scila.lu(linearTerm)
        diagu = np.nonzero( abs(np.diag(u))<1e-12 )
        if len(diagu[0])>0:
          raise ValueError("The map might not be increasing")
        sign_log_det , log_det =  np.linalg.slogdet(linearTerm)
        if sign_log_det<0:
          raise ValueError("The map should be monotone increasing")
        self.constantTerm = np.copy(constantTerm)
        self.linearTerm = np.copy(linearTerm)
        self.log_det_linearTerm = log_det

    def evaluate(self, x, *args, **kwargs):
        r""" Evaluate the transport map at the points :math:`{\bf x} \in \mathbb{R}^{m \times d}`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- transformed points

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        #Implementation valid also for sparse linear terms
        out = np.transpose( self.linearTerm.dot( x.transpose() ) ) + self.constantTerm 
        #out =  np.dot( x,self.linearTerm.transpose() ) + self.constantTerm
        return out

    def grad_x(self, x, *args, **kwargs):
        r""" Compute :math:`\nabla_{\bf x} \hat{T}({\bf x},{\bf a})`.

         .. math::
            :nowrap:

            \nabla_{\bf x} \hat{T}({\bf x},{\bf a}) =
                 \begin{bmatrix}
                 \nabla_{\bf x}  \hat{T}_1({\bf x},{\bf a})  \\
                 \nabla_{\bf x}  \hat{T}_2({\bf x},{\bf a})  \\
                 \vdots \\
                 \nabla_{\bf x}  \hat{T}_d({\bf x},{\bf a})
                 \end{bmatrix}

         for every evaluation point.

        Args:
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]) --
           gradient matrix (constant at every evaluation point).

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """

        if scisp.issparse(self.linearTerm):
            grad = self.linearTerm.toarray()
        else:
            grad = self.linearTerm

        return grad[np.newaxis,:]

    def hess_x(self, x, *args, **kwargs):
        r""" Compute :math:`\nabla^2_{\bf x} \hat{T}({\bf x},{\bf a})`.

            Args:
            x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

            Returns:
            (:class:`ndarray<numpy.ndarray>` [:math:`d,d,d`]) --
            Hessian matrix (zero everywhere).

            Raises:
            ValueError: if :math:`d` does not match the dimension of the transport map.
            """
        return np.zeros((self.dim, self.dim, self.dim))

    def grad_x_inverse(self, x, *args, **kwargs):
        r""" Compute :math:`\nabla_{\bf x} \hat{T}^{-1}({\bf x},{\bf a})`.

        Args:
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]) --
           gradient matrix (constant at every evaluation point).

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """
        if np.allclose(self.linearTerm, np.tril(self.linearTerm)):
            return scila.solve_triangular(self.linearTerm, np.eye(self.dim),
                                          lower=True)
        else:
            raise NotImplementedError("Not implemented at this time...")

    def hess_x_inverse(self, x, *args, **kwargs):
        r""" Compute :math:`\nabla^2_{\bf x} \hat{T}^{-1}({\bf x},{\bf a})`.

        Args:
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`d,d,d`]) --
           Hessian matrix (zero everywhere).

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """
        return np.zeros((self.dim, self.dim, self.dim))

    def log_det_grad_x(self, x, *args, **kwargs):
        r""" Compute: :math:`\log \det \nabla_{\bf x} \hat{T}({\bf x}, {\bf a})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`float`) --
           :math:`\log \det \nabla_{\bf x} \hat{T}({\bf x}, {\bf a})`
           (constant at every evaluation point)

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        return self.log_det_linearTerm

    def det_grad_x(self, x, *args, **kwargs):
        r""" Compute: :math:`\det \nabla_{\bf x} \hat{T}({\bf x}, {\bf a})`.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`float`) --
           :math:`\det \nabla_{\bf x} \hat{T}({\bf x}, {\bf a})`
           (constant at every evaluation point)

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.
        """
        return np.exp(self.log_det_linearTerm)

    def grad_x_log_det_grad_x(self, x, *args, **kwargs):
        r""" Compute: :math:`\nabla_{\bf x} \log \det \nabla_{\bf x} T({\bf x}, {\bf a})`

        Args:
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
           precomp (:class:`dict<dict>`): dictionary of precomputed values

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
           :math:`\nabla_{\bf x} \log \det \nabla_{\bf x} T({\bf x}, {\bf a})`
           at every evaluation point

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.

        .. seealso:: :func:`log_det_grad_x`.
        """
        return np.zeros((x.shape[0],self.dim))

    def log_det_grad_x_inverse(self, x, *args, **kwargs):
        r""" Compute: :math:`\log \det \nabla_{\bf x} T^{-1}({\bf x}, {\bf a})`.

        Args:
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (float) --
           :math:`\log \det \nabla_{\bf x} T^{-1}({\bf x}, {\bf a})`
           (constant at every evaluation point)
        """
        return - self.log_det_linearTerm

    def grad_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        return np.zeros((x.shape[0],self.dim))

    def hess_x_log_det_grad_x_inverse(self, x, *args, **kwargs):
        return np.zeros((x.shape[0],self.dim,self.dim))

    def inverse(self, y, *args, **kwargs):
        r""" Compute: :math:`\hat{T}^{-1}({\bf y},{\bf a})`

        Args:
           y (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
           :math:`\hat{T}^{-1}({\bf y},{\bf a})` for every evaluation point
        """
        if y.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = npla.solve(self.linearTerm, (y - self.constantTerm).T).T
        return out

    def grad_x_log_pullback(self, pi, x, params_t=None, params_pi=None, idxs_slice=slice(None)):
        r""" Compute: :math:`\nabla_{\bf x}\left[ \log \pi \circ \hat{T}({\bf x,a}) + \log \vert\det \nabla_{\bf x}\hat{T}({\bf x,a})\vert \right]`

        Args:
           pi (:class:`Distributions.Distribution`): distribution to be pulled back
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
           params_t (dict): parameters for the evaluation of :math:`T_{\bf a}`
           params_pi (dict): parameters for the evaluation of :math:`\pi`

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
           :math:`\nabla_{\bf x}\left[ \log \pi \circ \hat{T}({\bf x,a}) + \log \vert\det \nabla_{\bf x}\hat{T}({\bf x,a})\vert \right]`
           at every evaluation point

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.

        .. seealso:: :func:`log_pullback`, :func:`grad_x` and :func:`grad_x_log_det_grad_x`.
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        out = np.einsum( '...i,...ij->...j',
                         pi.grad_x_log_pdf(self.evaluate(x, params_t, idxs_slice), params_pi),
                         self.grad_x(x, params_t, idxs_slice) )
        return  out

    def hess_x_log_pullback(self, pi, x, params_t=None, params_pi=None, idxs_slice=slice(None)):
        r""" Compute :math:`\nabla^2_{\bf x}\left[ \log \pi \circ \hat{T}({\bf x,a}) + \log \vert\det \nabla_{\bf x}\hat{T}({\bf x,a})\vert \right]`.

        Args:
           pi (:class:`Distributions.Distribution`): distribution to be pulled back
           x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
           params_t (dict): parameters for the evaluation of :math:`T_{\bf a}`
           params_pi (dict): parameters for the evaluation of :math:`\pi`

        Returns:
           (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
           :math:`\nabla^2_{\bf x}\left[ \log \pi \circ \hat{T}({\bf x,a}) + \log \vert\det \nabla_{\bf x}\hat{T}({\bf x,a})\vert \right]`
           at every evaluation point

        Raises:
           ValueError: if :math:`d` does not match the dimension of the transport map.

        .. seealso:: :func:`grad_x_log_pullback`, :func:`log_pullback`, :func:`grad_x`, :func:`hess_x` and :func:`hess_x_log_det_grad_x`.
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        dx2logpi = pi.hess_x_log_pdf(
            self.evaluate(x, params_t, idxs_slice), params_pi )
        dxT = self.grad_x(x, params_t, idxs_slice)
        A = np.einsum('...kl,...lj->...kj',dx2logpi, dxT) #Hess logpi * grad_T
        out = np.einsum('...lk,...lj->...kj', dxT,A)
        return out



#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import sys, getopt
import warnings
import time
import matplotlib.pyplot as plt
import numpy as np
import numpy.random as npr
import scipy.stats as stats
from tabulate import tabulate

import TransportMaps as TM
import TransportMaps.Distributions as DIST
import TransportMaps.tests.TestFunctions as TF

logging.basicConfig(level=logging.DEBUG)

# TESTS
#  0: 1D - N(mu,sigma)
#  NOT WORKING 1: 1D - Mixture X~N(-1,1), Y~N(1,0.5), p=0.3
#  2: 1D - T(x) = a + b*x + c * arctan(d + e*x)
#  3: 1D - Y = T(x) = exp(a*x) -- Corresponding to Y~logN(0,a^2)
#  4: 1D - Y~Logistic(\mu,s)
#  5: 1D - Y~Gamma(loc,scale)
#  6: 1D - Y~Beta(a,b)
#  7: 1D - Gumbel distribution
#  NOTWORKING 8: Bayesian
#  9: 2D - N(mu,sigma)
# 10: 2D - Banana
AVAIL_TESTN = {0: '1D Std. Normal',
               2: '1D ArcTan',
               3: '1D LogNormal',
               4: '1D Logistic',
               5: '1D Gamma',
               6: '1D Beta',
               7: '1D Gumbel',
               9: '2D Std. Normal',
               10: '2D Banana'}

# SPAN APPROXIMATION
# 'full': Full order approximation
# 'total': Total order approximation
AVAIL_SPAN = ['full', 'total']

# QUADRATURES
# 0: Monte Carlo
# 3: Gauss quadrature
AVAIL_QUADRATURE = {0: 'Monte Carlo',
                    3: 'Gauss quadrature'}

def usage():
    print('KLdivergence-InverseMap-DirectConstruction.py -t <test_number> -s <span_type> -q <quad_type> -n <quad_n_points> [--tol=<tolerance> --with-reg=<alpha> --ord-list=<comma_sep_order_list> --no-plotting]')

def print_avail_span():
    print('Available <span_type>: %s' % AVAIL_SPAN)

def print_avail_qtype():
    print('Available <quad_type>:')
    for qtypen, qtypename in AVAIL_QUADRATURE.items():
        print('  %d: %s' % (qtypen, qtypename))

def print_avail_testn():
    print('Available <test_number>:')
    for testn, testname in AVAIL_TESTN.items():
        print('  %d: %s' % (testn, testname))

def print_example():
    print('KLdivergence-InverseMap-DirectConstruction.py -t 2 -s total -q 3 -n 30 --tol=1e-4 --with-reg=1e-3 --ord-list=1,3,5,7,9')
        
def full_usage():
    usage()
    print_avail_testn()
    print_avail_span()
    print_avail_qtype()
    print_example()

argv = sys.argv[1:]

PLOTTING = True
TESTN = None
SPAN = None
QTYPE = None
QNUM = None
REG = None
# Orders
ORDERS = [1,3,5]
# Tolerance
TOL = 1e-5
try:
    opts, args = getopt.getopt(argv,"ht:s:q:n:",["testn=", "span_type=",
                                                 "qtype=", "qnum=",
                                                 "tol=", "with-reg=", "ord-list=",
                                                 "no-plotting"])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt in ("-t", "--testn"):
        TESTN = int(arg)
        if TESTN not in AVAIL_TESTN:
            usage()
            print_avail_testn()
            sys.exit(1)
    elif opt in ("-s", "--span_type"):
        SPAN = arg
        if SPAN not in AVAIL_SPAN:
            usage()
            print_avail_span()
            sys.exit(1)
    elif opt in ("-q", "--qtype"):
        QTYPE = int(arg)
        if QTYPE not in AVAIL_QUADRATURE:
            usage()
            print_avail_qtype()
            sys.exit(1)
    elif opt in ("-n", "--qnum"):
        QNUM = int(arg)
    elif opt in ("--tol"):
        TOL = float(arg)
    elif opt in ("--with-reg"):
        REG = {'type': 'L2',
               'alpha': float(arg)}
    elif opt in ("--ord-list"):
        ORDERS = [int(s) for s in arg.split(',')]
    elif opt in ("--no-plotting"):
        PLOTTING = False
if None in [TESTN, SPAN, QTYPE, QNUM]:
    full_usage()
    sys.exit(3)

# PLOTTING
fsize = (6,5)
discr = 101

# Get test title and parameters
title, setup, test = TF.get(TESTN)

# # REGULARIZATION
# # None: No regularization
# # L2: L2 regularization
# REG = None
# # REG = {'type': 'None',
# #        'alpha': 1e-3}
if REG is None:
    tit_reg = 'None'
elif REG['type'] == 'L2':
    tit_reg = REG['type'] + '-' + str(REG['alpha'])

# L2 error estimation
# 3: Gauss quadrature of order n
# 0: Monte Carlo quadrature with n point
qtype = QTYPE
if qtype == 0:
    qparams = QNUM
    tit_intest = "MC" + '-' + str(qparams)
elif qtype == 3:
    qparams = [QNUM] * setup['dim']
    tit_intest = "Quad" + '-' + str(qparams[0])

# Gradient information
# 0: derivative free
# 1: gradient information
# 2: Hessian information
ders = 1

# MC samples
N_samp = 10000

# Generate samples of reference and target
ref_samp = test['base_distribution'].rvs(N_samp)
T_samp = test['support_map']( test['target_map'](ref_samp) )

# PLOTTING PRE-COMPUTATION
# Start plotting of the 1d map
if PLOTTING:
    pspan = setup['plotspan']
    rspan = setup['refplotspan']

    # Plot the 1d map
    X1d = np.vstack( (np.linspace(rspan[0,0],rspan[0,1],discr),
                      np.zeros((setup['dim']-1,discr))) ).T
    t1 = test['support_map']( test['target_map'](X1d) )[:,0]
    fig_map_1d = plt.figure()
    ax_map_1d = fig_map_1d.add_subplot(1,1,1)
    ax_map_1d.plot(X1d[:,0], t1, 'k--', linewidth=2, label='exact')

    # Plot the 1d distribution
    fig_kde_1d = plt.figure()
    ax_kde_1d = fig_kde_1d.add_subplot(1,1,1)
    # Exact distribution 1d
    X1d_kde = np.linspace(pspan[0,0],pspan[0,1],discr).reshape((discr,1))
    if setup['dim'] == 1:
        # Plot analytic
        pdf1d = test['target_distribution'].pdf(X1d_kde)
        ax_kde_1d.plot(X1d_kde, pdf1d, 'k--', linewidth=2, label='$\pi^1_{tar}$')
    else:
        # Gaussian KDE approximation
        T1_kde = stats.gaussian_kde(T_samp[:,0][:,np.newaxis].T)
        pdf1d = T1_kde(X1d_kde.T)
        ax_kde_1d.plot(X1d_kde, pdf1d, 'k--', linewidth=2, label='kde-$\pi^1_{tar}$')

    if setup['dim'] >= 2:
        # Plot the 2d transport map
        x = np.linspace(rspan[0,0],rspan[0,1],discr)
        y = np.linspace(rspan[1,0],rspan[1,1],discr)
        xx,yy = np.meshgrid(x,y)
        X2d = np.vstack( (xx.flatten(),yy.flatten(),np.zeros((setup['dim']-2,discr**2))) ).T
        t2 = test['support_map']( test['target_map'](X2d) )[:,1]
        fig_maps2d = []
        ax_maps = []
        levels_maps2d = np.linspace(np.min(t2),np.max(t2),20)
        for i,order in enumerate(ORDERS):
            fig_maps2d.append( plt.figure() )
            ax_map = fig_maps2d[-1].add_subplot(1,1,1)
            ax_map.contour(xx,yy,t2.reshape(xx.shape),
                           levels=levels_maps2d)
            ax_maps.append( ax_map )
            plt.title("Order %d" % order)

        # Plot PDF 2d
        fig_pdf2d = []
        axs_pdf2d = []
        x = np.linspace(pspan[0,0],pspan[0,1],discr)
        y = np.linspace(pspan[1,0],pspan[1,1],discr)
        xx2d,yy2d = np.meshgrid(x,y)
        X2d_kde = np.vstack( (xx2d.flatten(),yy2d.flatten(),
                              np.zeros((setup['dim']-2,discr**2))) ).T
        if setup['dim'] == 2:
            pdf2d = test['target_distribution'].pdf(X2d_kde)
        else:
            T2_kde = stats.gaussian_kde(T_samp[:,:1].T)
            pdf2d = T2_kde(X2d_kde.T)
        levels_pdf2d = np.linspace(np.min(pdf2d),np.max(pdf2d),20)
        for i,order in enumerate(ORDERS):
            fig_pdf2d.append( plt.figure() )
            ax_pdf2d = fig_pdf2d[-1].add_subplot(1,1,1)
            ax_pdf2d.contour(xx2d,yy2d,pdf2d.reshape(xx2d.shape),
                             levels=levels_pdf2d)
            axs_pdf2d.append( ax_pdf2d )
            plt.title("Order %d" % order)

# Define header for the output log
tit_type = 'monot'
if ders in [0,1]:
    log_header = ['Order','#F.ev.','#J.ev','$L^2$-err']
else:
    log_header = ['Order','#F.ev.','#J.ev','#H.ev','$L^2$-err']

log = [ ]

for i_ord,order in enumerate(ORDERS):
    tmp_coeffs = None
    for nest_ord in range(1,order+1):
        print("\nOrder: %d" % nest_ord)
        # Build the transport map (isotropic for each entry)
        tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
            setup['dim'], nest_ord, SPAN)
        # Construct distribution
        # Distribution T^\sharp \pi
        tm_distribution = DIST.PullBackTransportMapDistribution(tm_approx, test['base_distribution'])
        # Target distribution to be approximated L^\sharp \pi_{\rm tar}
        target_distribution = DIST.PullBackTransportMapDistribution(test['support_map'],
                                                          test['target_distribution'])
        if tmp_coeffs is None:
            x0 = None
        else:
            x0 = tm_approx.get_coeffs()
            x0[:nest_ord] = tmp_coeffs
        # SOLVE min D_KL (T^\sharp \pi || L^\sharp \pi_{\rm tar})
        log_entry_solve = tm_distribution.minimize_kl_divergence(target_distribution, qtype=qtype,
                                                            qparams=qparams,
                                                            x0=x0,
                                                            regularization=REG,
                                                            tol=TOL, ders=ders,
                                                            batch_size=[None,None,20])
        tmp_coeffs = tm_approx.get_coeffs()

    log_entry = [order] + log_entry_solve

    # Construct approximate distribution L_\sharp T^\sharp \pi
    approx_distribution = DIST.PushForwardTransportMapDistribution(test['support_map'], tm_distribution)

    # Push forward reference samples in two steps
    # (1) Map x from \pi to y from T^\sharp \pi
    # (2) Map y from T^\sharp \pi to z from L_\sharp T^\sharp \pi
    tmp = tm_distribution.map_samples_base_to_target( ref_samp )
    fapp_samp = approx_distribution.map_samples_base_to_target( tmp ) 

    # PLOT
    if PLOTTING:
        # Map to T^\sharp \pi
        t1_approx_tmp = tm_distribution.map_samples_base_to_target(X1d)
        # Map to L_\sharp T^\sharp \pi
        t1_approx = approx_distribution.map_samples_base_to_target(t1_approx_tmp)[:,0]
        # Plot 1d transport map
        ax_map_1d.plot(X1d[:,0], t1_approx, label='ord %d' % order)

        # Plot the 1d distribution
        if setup['dim'] == 1:
            pdf1d_approx = approx_distribution.pdf(X1d_kde)
        else:
            t1_kde_approx = stats.gaussian_kde(fapp_samp[:,0][:,np.newaxis].T)
            pdf1d_approx = t1_kde_approx(X1d_kde.T)
        ax_kde_1d.plot(X1d_kde, pdf1d_approx, label='ord %d' % order)


        if setup['dim'] >= 2:
            # Map to T^\sharp \pi
            t2_approx_tmp = tm_distribution.map_samples_base_to_target(X2d)
            # Map to L_\sharp T^\sharp \pi
            t2_approx = approx_distribution.map_samples_base_to_target(t2_approx_tmp)[:,1]
            # Plot 2d transport map
            ax_maps[i_ord].contour(xx2d, yy2d, t2_approx.reshape(xx2d.shape),
                                   linestyles='dashed', levels=levels_maps2d)

            # Plot the 2d distribution
            if setup['dim'] == 2:
                pdf2d_approx = approx_distribution.pdf(X2d_kde)
            else:
                t2_kde_approx = stats.gaussian_kde(fapp_samp[:,:1].T)
                pdf2d_approx = t2_kde_approx(X2d_kde.T)
            axs_pdf2d[i_ord].contour(xx2d,yy2d,pdf2d_approx.reshape(xx2d.shape),
                                     linestyles='dashed', levels=levels_pdf2d)

    # Compute L2 error
    dist = np.sqrt( np.sum((T_samp - fapp_samp)**2., axis=1) )
    log_entry.append( np.sqrt( np.sum( dist**2. ) / float(N_samp) ) )

    log.append( log_entry )

if PLOTTING:
    fig_map_1d.gca().legend()
    fig_kde_1d.gca().legend()
    plt.show(False)

# # Decorate map plotting
# # ax_map.set_ylim([f(X[0]), f(X[-1])])
# ax_map.scatter(x, np.zeros(x.shape))
# ax_map.set_xlabel('x')
# ax_map.set_ylabel('$T(x)$')
# # ax_map.set_ylabel('$ (F^{-1} \circ F_{ref})x $')
# # ax_map.set_ylim([np.min(exact_tm),np.max(exact_tm)])
# ax_map.legend(loc='best')
# ax_map.grid(True)

# # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.pdf' % \
# #                 (title, tit_type, tit_reg, tit_intest))
# # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.eps' % \
# #                 (title, tit_type, tit_reg, tit_intest))

# # Decorate kde plotting
# ax_kde.set_xlabel('x')
# ax_kde.set_ylabel('PDF')
# ax_kde.legend(loc='best')

# # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.pdf' % \
# #                 (title, tit_type, tit_reg, tit_intest))
# # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.eps' % \
# #                 (title, tit_type, tit_reg, tit_intest))

# plt.show(False)

print(" ")
print(tabulate(log, headers=log_header))
print(" ")

print(tabulate(log, headers=log_header, tablefmt="latex", floatfmt=".2e"))


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import scipy.special
from TransportMaps.Distributions.DistributionBase import Distribution

__all__ = ['BODjoint']

class BODjoint(Distribution):
    def __init__(self, numY, sigma = np.sqrt(1e-3) ):
        timeY=np.arange(numY)+1.
        dimDistribution = numY+2
        super(BODjoint,self).__init__(dimDistribution)
        self.numY = numY
        self.timeY  = timeY
        self.sigma = sigma
    def pdf(self, x, params=None):
        return np.exp(self.log_pdf(x, params))
    def log_pdf(self, x, params=None):
        # x is a 2-d array of points. The first dimension corresponds to the number of points.
        #The first numY columns of x refer to the data
        numY = self.numY
        Y = x[:,0:numY]
        theta1=x[:,numY] #The last two components refer to the parameters
        theta2=x[:,numY+1]
        a = .4  + .4*( 1 + scipy.special.erf( theta1/np.sqrt(2) )  )
        b = .01 + .15*( 1 + scipy.special.erf( theta2/np.sqrt(2) )  )
        return -1.0/(2*self.sigma**2) * \
            np.sum( (Y - a[:,np.newaxis] * \
                     ( 1 - np.exp( -np.outer(b, self.timeY) ) ) )**2 , axis=1) + \
            -.5*( theta1**2 + theta2**2 )
    def grad_x_log_pdf(self, x, params=None):
        numY = self.numY
        Y = x[:,0:numY]
        theta1=x[:,numY] #The last two components refer to the parameters
        theta2=x[:,numY+1]
        a = .4  + .4*( 1 + scipy.special.erf( theta1/np.sqrt(2) )  )
        b = .01 + .15*( 1 + scipy.special.erf( theta2/np.sqrt(2) )  )
        da_theta1 = .4*np.sqrt(2/np.pi)*np.exp( -theta1**2/2.)
        db_theta2 = .15*np.sqrt(2/np.pi)*np.exp( -theta2**2/2.)
        grad = np.zeros( x.shape )
        for jj in np.arange(numY):
            grad[:,numY] -= -da_theta1/(self.sigma**2) * \
                            ( 1 - np.exp( - b*self.timeY[jj])  ) * \
                            ( Y[:,jj] - a*(1- np.exp(-b*self.timeY[jj])) )
            grad[:,numY+1] -= -1.0/(self.sigma**2)*self.timeY[jj]*db_theta2*a * \
                              np.exp(-b*self.timeY[jj]) * \
                              ( Y[:,jj] - a*(1- np.exp(-b*self.timeY[jj])) )
            grad[:,jj] = -1.0/self.sigma**2 * \
                         ( Y[:,jj] - a*(1-np.exp(-b*self.timeY[jj])) )
        grad[:,-2] -=  theta1
        grad[:,-1] -=  theta2
        return grad
    def hess_x_log_pdf(self, x, params=None):
        numY = self.numY
        Y = x[:,0:numY]
        theta1=x[:,numY] #The last two components refer to the parameters
        theta2=x[:,numY+1]
        a = .4  + .4*( 1 + scipy.special.erf( theta1/np.sqrt(2) )  )
        b = .01 + .15*( 1 + scipy.special.erf( theta2/np.sqrt(2) )  )
        da_theta1 = .4*np.sqrt(2/np.pi)*np.exp( -theta1**2/2.)
        db_theta2 = .15*np.sqrt(2/np.pi)*np.exp( -theta2**2/2.)
        d2a_theta1 = -theta1*da_theta1
        d2b_theta2 = -theta2*db_theta2
        Hess_x = np.zeros( (x.shape[0], x.shape[1],  x.shape[1]) )
        for jj in np.arange(numY):
            Hess_x[:,numY,numY]-= \
                -(1-np.exp(-b*self.timeY[jj]))/(self.sigma**2) * \
                ( d2a_theta1*( Y[:,jj] - a*(1- np.exp(-b*self.timeY[jj])) ) - \
                  da_theta1**2*(1-np.exp(-b*self.timeY[jj]))  )
            Hess_x[:,numY+1,numY]-= \
                -da_theta1/(self.sigma**2) * \
                ( db_theta2*self.timeY[jj]*np.exp(-b*self.timeY[jj]) * \
                  (Y[:,jj]-a+a*np.exp(-b*self.timeY[jj])) + \
                  (1-np.exp(-b*self.timeY[jj])) * \
                  (-a*self.timeY[jj]*db_theta2*np.exp(-b*self.timeY[jj])))
            Hess_x[:,numY+1,numY+1]-= \
                -self.timeY[jj]*a/(self.sigma**2) * \
                np.exp(-b*self.timeY[jj]) * \
                ( ( Y[:,jj] - a*(1- np.exp(-b*self.timeY[jj])) ) * \
                  ( d2b_theta2 - self.timeY[jj]*db_theta2**2 ) - \
                  db_theta2**2 *self.timeY[jj]*a*np.exp(-b*self.timeY[jj]))
            Hess_x[:,numY,jj] = da_theta1/(self.sigma**2)*(1-np.exp(-b*self.timeY[jj]))
            Hess_x[:,numY+1,jj] = 1/(self.sigma**2)*self.timeY[jj] * \
                                  db_theta2*a*np.exp(-b*self.timeY[jj])
            Hess_x[:,jj, numY] = Hess_x[:,numY,jj]
            Hess_x[:,jj, numY+1] = Hess_x[:,numY+1, jj]
        Hess_x[:,numY, numY+1] = Hess_x[:,numY+1, numY]
        Hess_x[:,numY,numY]-=1
        Hess_x[:,numY+1,numY+1]-=1
        for kk in np.arange(x.shape[0]):
            for jj in np.arange(numY):
                Hess_x[ kk , jj, jj]  = -1/(self.sigma**2)
        return Hess_x


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import SpectralToolbox.Spectral1D as S1D
import TransportMaps as TM
import TransportMaps.Functionals as FUNC
import TransportMaps.Maps as MAPS

__all__ = ['Default_IsotropicIntegratedExponentialTriangularTransportMap',
           'Default_IsotropicIntegratedExponentialDiagonalTransportMap',
           'Default_IsotropicIntegratedSquaredTriangularTransportMap',
           'Default_IsotropicIntegratedSquaredDiagonalTransportMap',
           'Default_IsotropicMonotonicLinearSpanTriangularTransportMap',
           'Default_IsotropicLinearSpanTriangularTransportMap',
           'Default_LinearSpanTriangularTransportMap']

def Default_IsotropicIntegratedExponentialTriangularTransportMap(
        dim, order, span='total', active_vars=None, btype='poly',
        common_basis_flag=True):
    r""" Generate a triangular transport map with default settings.

    Args:
      dim (int): dimension :math:`d` of the map
      order (int): isotropic order of the map
      span (str): 'full' for full order approximations, 'total' for total order
        approximations. If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the approximation type fore each component
        :math:`T^{(k)}`.
      active_vars (list): list of :math:`d` lists containing the row sparsity
        pattern of the transport, i.e. the active variables for each component.
        Default ``None`` will generate a full triangular map.
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.
      common_basis_flag (bool): use acceleration provided by common basis among the
        components (``span`` and ``btype`` must be a string or a list with all equal
        elements).

    Returns:
      (:class:`IntegratedExponentialTriangularTransportMap<IntegratedExponentialTriangularTransportMap>`)
      -- the constructed transport map
    """
    # Initialize the list of components
    approx_list = []
    # Initialize the list of active variables
    if active_vars is None:
        active_vars = [ list(range(i+1)) for i in range(dim) ]
    else:
        if len(active_vars) != dim:
            raise ValueError("List of active variables must be dim long.")
        for d, avars in enumerate(active_vars):
            if sorted(avars) != avars:
                raise ValueError("List of active components must be provided in " + \
                                 "sorted order.")
            if avars[-1] != d:
                raise ValueError("List of active components must include at least" + \
                                 "the diagonal term.")
    # Initialize the span type
    if isinstance(span,str):
        span_list = [span] * dim
    else:
        if len(span) != dim:
            raise ValueError("List of span types must be dim long.")
        else:
            span_list = span
    # Initialize the basis type
    if isinstance(btype, str):
        btype_list = [btype] * dim
    else:
        if len(btype) != dim:
            raise AttributeError("List of basis types must be dim long.")
        else:
            btype_list = btype
    # Check whether it is possible to use common basis
    common_basis_flag = common_basis_flag and all(x==span_list[0] for x in span_list) and \
                        all(x==btype_list[0] for x in btype_list)

    # Prepare basis in case common_basis
    if common_basis_flag:
        if btype_list[0] == 'poly':
            c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim)]
            e_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim-1)] + \
                           [S1D.ConstantExtendedHermiteProbabilistsFunction()]
            # e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction()
            #                 for i in range(dim)]
        elif btype_list[0] == 'rbf':
            if span_list[0] != 'full':
                raise ValueError("The basis span must be 'full' for basis type 'rbf'")
            c_basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim)]
            c_basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim-1)] + \
                [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)]
            # e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
            #     order, 0.9) for i in range(dim)]
        else:
            raise ValueError("Input btype is invalid.")

    # Instantiate the components
    for d, (avars, span, btype) in enumerate(zip(active_vars, span_list, btype_list)):
        c_orders_list = [order]*(len(avars)-1)+[0]
        if common_basis_flag:
            c_basis_comp = [c_basis_list[a] for a in avars]
            e_basis_comp = [e_basis_list[a] for a in avars]
        else:
            if btype == 'poly':
                c_basis_comp = [S1D.HermiteProbabilistsPolynomial()] * len(avars)
                c_basis_comp = [S1D.HermiteProbabilistsPolynomial()] * (len(avars)-1) + \
                               [S1D.ConstantExtendedHermiteProbabilistsFunction()]
                # e_basis_comp = [S1D.ConstantExtendedHermiteProbabilistsFunction()] * \
                #                len(avars)
            elif btype == 'rbf':
                if span != 'full':
                    raise ValueError("The basis span must be 'full' for basis type 'rbf'")
                c_basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * len(avars)
                e_basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * (len(avars)-1) + \
                    [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                        order, 0.9)]
                # e_basis_comp = [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                #     order, 0.9)] * len(avars)
                
        c_approx = FUNC.LinearSpanApproximation(
            c_basis_comp, spantype=span, order_list=c_orders_list)
        e_orders_list = [order-1]*len(avars)
        e_approx = FUNC.LinearSpanApproximation(
            e_basis_comp, spantype=span, order_list=e_orders_list)
        approx = FUNC.MonotonicIntegratedExponentialApproximation(c_approx, e_approx)
        approx_list.append( approx )

    # Instantiate the map
    if common_basis_flag:    
        tm_approx = MAPS.CommonBasisIntegratedExponentialTriangularTransportMap(
            active_vars, approx_list)
    else:
        tm_approx = MAPS.IntegratedExponentialTriangularTransportMap(
            active_vars, approx_list)
    return tm_approx

def Default_IsotropicIntegratedExponentialDiagonalTransportMap(
        dim, order, btype='poly'):
    r""" Generate a diagonal transport map with default settings.

    Args:
      dim (int): dimension :math:`d` of the map
      order (int): isotropic order of the map
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.

    Returns:
      (:class:`IntegratedExponentialTriangularTransportMap<IntegratedExponentialTriangularTransportMap>`)
      -- the constructed transport map
    """
    active_vars = [[d] for d in range(dim)]
    return Default_IsotropicIntegratedExponentialTriangularTransportMap(
        dim, order, span='total', active_vars=active_vars, btype=btype,
        common_basis_flag=False)

def Default_IsotropicIntegratedSquaredTriangularTransportMap(
        dim, order, span='total', active_vars=None, btype='poly',
        common_basis_flag=False):
    r""" Generate a triangular transport map with default settings.

    Args:
      dim (int): dimension :math:`d` of the map
      order (int): isotropic order of the map
      span (str): 'full' for full order approximations, 'total' for total order
        approximations. If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the approximation type fore each component
        :math:`T^{(k)}`.
      active_vars (list): list of :math:`d` lists containing the row sparsity
        pattern of the transport, i.e. the active variables for each component.
        Default ``None`` will generate a full triangular map.
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.
      common_basis_flag (bool): use acceleration provided by common basis among the
        components (``span`` and ``btype`` must be a string or a list with all equal
        elements).

    Returns:
      (:class:`IntegratedSquaredTriangularTransportMap<IntegratedSquaredTriangularTransportMap>`)
      -- the constructed transport map
    """
    # Initialize the list of components
    approx_list = []
    # Initialize the list of active variables
    if active_vars is None:
        active_vars = [ list(range(i+1)) for i in range(dim) ]
    else:
        if len(active_vars) != dim:
            raise ValueError("List of active variables must be dim long.")
        for d, avars in enumerate(active_vars):
            if sorted(avars) != avars:
                raise ValueError("List of active components must be provided in " + \
                                 "sorted order.")
            if avars[-1] != d:
                raise ValueError("List of active components must include at least" + \
                                 "the diagonal term.")
    # Initialize the span type
    if isinstance(span,str):
        span_list = [span] * dim
    else:
        if len(span) != dim:
            raise ValueError("List of span types must be dim long.")
        else:
            span_list = span
    # Initialize the basis type
    if isinstance(btype, str):
        btype_list = [btype] * dim
    else:
        if len(btype) != dim:
            raise AttributeError("List of basis types must be dim long.")
        else:
            btype_list = btype
    # Check whether it is possible to use common basis
    common_basis_flag = common_basis_flag and all(x==span_list[0] for x in span_list) and \
                        all(x==btype_list[0] for x in btype_list)

    # Prepare basis in case common_basis
    if common_basis_flag:
        if btype_list[0] == 'poly':
            c_basis_list = [S1D.LinearExtendedHermiteProbabilistsFunction() for i in range(dim)]
            # c_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim)]
            # c_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction() for i in range(dim)]
            e_basis_list = [S1D.LinearExtendedHermiteProbabilistsFunction()
                            for i in range(dim-1)] + \
                           [S1D.ConstantExtendedHermiteProbabilistsFunction()]
            # e_basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim-1)] + \
            #                [S1D.ConstantExtendedHermiteProbabilistsFunction()]
            # e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsFunction() for i in range(dim)]
        elif btype_list[0] == 'rbf':
            if span[0] != 'full':
                raise ValueError("The basis span must be 'full' for basis type 'rbf'")
            c_basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim)]
            e_basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim-1)] + \
                [ S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9) ]
            # e_basis_list = [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
            #     order, 0.9) for i in range(dim)]
        else:
            raise ValueError("Input btype is invalid.")

    # Instantiate the components
    for d, (avars, span, btype) in enumerate(zip(active_vars, span_list, btype_list)):
        c_orders_list = [order]*(len(avars)-1)+[0]
        if common_basis_flag:
            c_basis_comp = [c_basis_list[a] for a in avars]
            e_basis_comp = [e_basis_list[a] for a in avars]
        else:
            if btype == 'poly':
                c_basis_comp = [S1D.LinearExtendedHermiteProbabilistsFunction()] * len(avars)
                # c_basis_comp = [S1D.HermiteProbabilistsPolynomial()] * len(avars)
                # c_basis_comp = [S1D.ConstantExtendedHermiteProbabilistsFunction()] * len(avars)
                e_basis_comp = [S1D.LinearExtendedHermiteProbabilistsFunction()] * \
                               (len(avars)-1) + \
                               [S1D.ConstantExtendedHermiteProbabilistsFunction()]
                # e_basis_comp = [S1D.HermiteProbabilistsPolynomial()] * (len(avars)-1) + \
                #                [S1D.ConstantExtendedHermiteProbabilistsFunction()]
                # e_basis_comp = [S1D.ConstantExtendedHermiteProbabilistsFunction()] * \
                #                len(avars)
            elif btype == 'rbf':
                if span != 'full':
                    raise ValueError("The basis span must be 'full' for basis type 'rbf'")
                c_basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * len(avars)
                e_basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * (len(avars)-1) + \
                    [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                        order, 0.9)]
                # e_basis_comp = [S1D.ConstantExtendedHermiteProbabilistsRadialBasisFunction(
                #     order, 0.9)] * len(avars)
                
        c_approx = FUNC.LinearSpanApproximation(
                c_basis_comp, spantype=span, order_list=c_orders_list)
        e_orders_list = [order-1]*len(avars)
        e_approx = FUNC.LinearSpanApproximation(
                e_basis_comp, spantype=span, order_list=e_orders_list)
        approx = FUNC.MonotonicIntegratedSquaredApproximation(c_approx, e_approx)
        approx_list.append( approx )

    # Instantiate the map
    if common_basis_flag:    
        tm_approx = MAPS.CommonBasisIntegratedSquaredTriangularTransportMap(
            active_vars, approx_list)
    else:
        tm_approx = MAPS.IntegratedSquaredTriangularTransportMap(
            active_vars, approx_list)
    return tm_approx

def Default_IsotropicIntegratedSquaredDiagonalTransportMap(
        dim, order, btype='poly'):
    r""" Generate a diagonal transport map with default settings.

    Args:
      dim (int): dimension :math:`d` of the map
      order (int): isotropic order of the map
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.

    Returns:
      (:class:`IntegratedSquaredTriangularTransportMap<IntegratedSquaredTriangularTransportMap>`)
      -- the constructed transport map
    """
    active_vars = [[d] for d in range(dim)]
    return Default_IsotropicIntegratedSquaredTriangularTransportMap(
        dim, order, span='total', active_vars=active_vars, btype=btype,
        common_basis_flag=False)

    
def Default_IsotropicMonotonicLinearSpanTriangularTransportMap(
        dim, order, span='total', active_vars=None, btype='poly',
        common_basis_flag=True):
    r""" Generate a triangular transport map with default settings.

    Args:
      dim (int): dimension of the map
      order (int): isotropic order of the map
      span (str): 'full' for full order approximations, 'total' for total order
        approximations. If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the approximation type fore each component
        :math:`T^{(k)}`.
      active_vars (list): list of :math:`d` lists containing the row sparsity
        pattern of the transport, i.e. the active variables for each component.
        Default ``None`` will generate a full triangular map.
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.
      common_basis_flag (bool): use acceleration provided by common basis among the
        components (``span`` and ``btype`` must be a string or a list with all equal
        elements).

    Returns:
      (:class:`LinearSpanTriangularTransportMap<LinearSpanTriangularTransportMap>`) -- the constructed transport map
    """
    # Initialize the list of components
    approx_list = []
    # Initialize the list of active variables
    if active_vars is None:
        active_vars = [ list(range(i+1)) for i in range(dim) ]
    else:
        if len(active_vars) != dim:
            raise ValueError("List of active variables must be dim long.")
        for d, avars in enumerate(active_vars):
            if sorted(avars) != avars:
                raise ValueError("List of active components must be provided in " + \
                                 "sorted order.")
            if avars[-1] != d:
                raise ValueError("List of active components must include at least" + \
                                 "the diagonal term.")
    # Initialize the span type
    if isinstance(span,str):
        span_list = [span] * dim
    else:
        if len(span) != dim:
            raise ValueError("List of span types must be dim long.")
        else:
            span_list = span
    # Initialize the basis type
    if isinstance(btype, str):
        btype_list = [btype] * dim
    else:
        if len(btype) != dim:
            raise AttributeError("List of basis types must be dim long.")
        else:
            btype_list = btype
    # Check whether it is possible to use common basis
    common_basis_flag = common_basis_flag and all(x==span_list[0] for x in span_list) and \
                        all(x==btype_list[0] for x in btype_list)

    # Prepare basis in case common_basis
    if common_basis_flag:
        if btype_list[0] == 'poly':
            basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim)]
        elif btype_list[0] == 'rbf':
            if span_list[0] != 'full':
                raise ValueError("The basis span must be 'full' for basis type 'rbf'")
            basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim)]
        else:
            raise ValueError("Input btype is invalid.")

    # Instantiate the components
    for d, (avars, span, btype) in enumerate(zip(active_vars, span_list, btype_list)):
        orders_list = [order]*len(avars)
        if common_basis_flag:
            basis_comp = [basis_list[a] for a in avars]
        else:
            if btype == 'poly':
                basis_comp = [S1D.HermiteProbabilistsPolynomial()] * len(avars)
            elif btype == 'rbf':
                if span != 'full':
                    raise ValueError("The basis span must be 'full' for basis type 'rbf'")
                basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * len(avars)
            else:
                raise ValueError("Input btype is invalid.")
                
        approx = FUNC.MonotonicLinearSpanApproximation(
            basis_comp, spantype=span, order_list=orders_list)
        approx_list.append( approx )

    # Instantiate the map
    if common_basis_flag:    
        tm_approx = MAPS.MonotonicCommonBasisLinearSpanTriangularTransportMap(
            active_vars, approx_list)
    else:
        tm_approx = MAPS.MonotonicLinearSpanTriangularTransportMap(
            active_vars, approx_list)
    return tm_approx

def Default_IsotropicLinearSpanTriangularTransportMap(
        dim, order, span='total', active_vars=None, btype='poly',
        common_basis_flag=True):
    r""" Generate a triangular transport map with default settings.

    Args:
      dim (int): dimension of the map
      order (int): isotropic order of the map
      span (str): 'full' for full order approximations, 'total' for total order
        approximations. If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the approximation type fore each component
        :math:`T^{(k)}`.
      active_vars (list): list of :math:`d` lists containing the row sparsity
        pattern of the transport, i.e. the active variables for each component.
        Default ``None`` will generate a full triangular map.
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.
      common_basis_flag (bool): use acceleration provided by common basis among the
        components (``span`` and ``btype`` must be a string or a list with all equal
        elements).

    Returns:
      (:class:`LinearSpanTriangularTransportMap<LinearSpanTriangularTransportMap>`) -- the constructed transport map
    """
    # Initialize the list of components
    approx_list = []
    # Initialize the list of active variables
    if active_vars is None:
        active_vars = [ list(range(i+1)) for i in range(dim) ]
    else:
        if len(active_vars) != dim:
            raise ValueError("List of active variables must be dim long.")
        for d, avars in enumerate(active_vars):
            if sorted(avars) != avars:
                raise ValueError("List of active components must be provided in " + \
                                 "sorted order.")
            if avars[-1] != d:
                raise ValueError("List of active components must include at least" + \
                                 "the diagonal term.")
    # Initialize the span type
    if isinstance(span,str):
        span_list = [span] * dim
    else:
        if len(span) != dim:
            raise ValueError("List of span types must be dim long.")
        else:
            span_list = span
    # Initialize the basis type
    if isinstance(btype, str):
        btype_list = [btype] * dim
    else:
        if len(btype) != dim:
            raise AttributeError("List of basis types must be dim long.")
        else:
            btype_list = btype
    # Check whether it is possible to use common basis
    common_basis_flag = common_basis_flag and all(x==span_list[0] for x in span_list) and \
                        all(x==btype_list[0] for x in btype_list)

    # Prepare basis in case common_basis
    if common_basis_flag:
        if btype_list[0] == 'poly':
            basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim)]
        elif btype_list[0] == 'rbf':
            if span_list[0] != 'full':
                raise ValueError("The basis span must be 'full' for basis type 'rbf'")
            basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim)]
        else:
            raise ValueError("Input btype is invalid.")

    # Instantiate the components
    for d, (avars, span, btype) in enumerate(zip(active_vars, span_list, btype_list)):
        orders_list = [order]*len(avars)
        if common_basis_flag:
            basis_comp = [basis_list[a] for a in avars]
        else:
            if btype == 'poly':
                basis_comp = [S1D.HermiteProbabilistsPolynomial()] * len(avars)
            elif btype == 'rbf':
                if span != 'full':
                    raise ValueError("The basis span must be 'full' for basis type 'rbf'")
                basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * len(avars)
            else:
                raise ValueError("Input btype is invalid.")
                
        approx = FUNC.LinearSpanApproximation(
            basis_comp, spantype=span, order_list=orders_list)
        approx_list.append( approx )

    # Instantiate the map
    if common_basis_flag:    
        tm_approx = MAPS.CommonBasisLinearSpanTriangularTransportMap(
            active_vars, approx_list)
    else:
        tm_approx = MAPS.LinearSpanTriangularTransportMap(
            active_vars, approx_list)
    return tm_approx

    
def Default_LinearSpanTriangularTransportMap(dim, midxs_list, active_vars,
                                             btype='poly', common_basis_flag=True):
    r""" Generate a linear span triangular transport map with default settings and user defined sparsity and orders.

    Args:
      dim (int): dimension of the map
      midxs_list (list): list of :math:`d` lists of multi-indices for each component
      active_vars (list): list of :math:`d` lists containing the row sparsity
        pattern of the transport, i.e. the active variables for each component.
      btype (string): ``poly`` uses Hermite functions,
        ``rbf`` uses radial basis functions.
        If a :class:`list<list>` of ``dim`` strings is provided,
        these will define the basis type fore each component
        :math:`T^{(k)}`.
      common_basis_flag (bool): use acceleration provided by common basis among the
        components (``btype`` must be a string or a list with all equal
        elements).

    Returns:
      (:class:`LinearSpanTriangularTransportMap<LinearSpanTriangularTransportMap>`) -- the constructed transport map
    """
    # Initialize the list of components
    approx_list = []
    # Initialize the list of active variables
    if len(active_vars) != dim:
        raise ValueError("List of active variables must be dim long.")
    for d, avars in enumerate(active_vars):
        if sorted(avars) != avars:
            raise ValueError("List of active components must be provided in " + \
                             "sorted order.")
        if avars[-1] != d:
            raise ValueError("List of active components must include at least" + \
                             "the diagonal term.")
    # Initialize the basis type
    if isinstance(btype, str):
        btype_list = [btype] * dim
    else:
        if len(btype) != dim:
            raise AttributeError("List of basis types must be dim long.")
        else:
            btype_list = btype
    # Check whether it is possible to use common basis
    common_basis_flag = common_basis_flag and all(x==btype_list[0] for x in btype_list)

    # Prepare basis in case common_basis
    if common_basis_flag:
        if btype_list[0] == 'poly':
            basis_list = [S1D.HermiteProbabilistsPolynomial() for i in range(dim)]
        elif btype_list[0] == 'rbf':
            basis_list = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                order, 0.9) for i in range(dim)]
        else:
            raise ValueError("Input btype is invalid.")

    # Instantiate the components
    for d, (avars, midxs, btype) in enumerate(zip(active_vars, midxs_list, btype_list)):
        if common_basis_flag:
            basis_comp = [basis_list[a] for a in avars]
        else:
            if btype == 'poly':
                basis_comp = [S1D.HermiteProbabilistsPolynomial()] * len(avars)
            elif btype == 'rbf':
                basis_comp = [S1D.LinearExtendedHermiteProbabilistsRadialBasisFunction(
                    order, 0.9)] * len(avars)
            else:
                raise ValueError("Input btype is invalid.")
        approx = FUNC.LinearSpanApproximation(basis_comp, midxs)
        approx_list.append( approx )

    # Instantiate the map
    if common_basis_flag:    
        tm_approx = MAPS.CommonBasisLinearSpanTriangularTransportMap(
            active_vars, approx_list)
    else:
        tm_approx = MAPS.LinearSpanTriangularTransportMap(
            active_vars, approx_list)
    return tm_approx


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2017 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest

try:
    import mpi_map
    MPI_SUPPORT = True
except:
    MPI_SUPPORT = False

class Scripts_unittest(unittest.TestCase):
    def setUp(self):
        import dill
        import numpy as np
        import numpy.random as npr
        import TransportMaps.Distributions as DIST
        # Parameters
        self.dim = 5
        self.dist_fname = 'Distribution.dill'
        self.fname_list = [self.dist_fname]
        # Build and store target distribution
        tar_mu = npr.randn(self.dim)
        tar_sig = npr.randn(self.dim**2).reshape((self.dim, self.dim))
        tar_sig2 = np.dot(tar_sig, tar_sig.T)
        target = DIST.GaussianDistribution(tar_mu, sigma=tar_sig2)
        with open(self.dist_fname, 'wb') as out_stream:
            dill.dump(target, out_stream)

    def tearDown(self):
        import os
        for fname in self.fname_list:
            if os.path.exists(fname):
                os.remove(fname)

    def postprocess(self):
        from distutils.spawn import find_executable
        from subprocess import call
        # Parameters
        self.post_fname = 'Postprocess.dill'
        self.fname_list.append( self.post_fname )
        self.fname_list.append( self.post_fname + '.bak' )
        self.fname_list.append( self.post_fname + '.hdf5' )
        # Find path to script (to avoid long paths with sh..)
        script_path = find_executable("tmap-postprocess")
        # Run post-process script with different options
        # Target aligned conditionals
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--aligned-conditionals=exact-target",
                       "--no-plotting",
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )
        # Random aligned conditionals
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--random-conditionals=approx-base", "--no-plotting",
                       "--rndc-n-points-x-ax=20", "--rndc-n-plots-x-ax=3",
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )
        # Pullback aligned conditionals
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--aligned-conditionals=approx-base", "--no-plotting",
                       "--alc-n-points-x-ax=20",
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )
        # Variance diagnostic (MC)
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--var-diag=exact-base",
                       "--var-diag-qtype=0", "--var-diag-qnum=100",
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )
        # Variance diagnostic (MC increased)
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--var-diag=exact-base",
                       "--var-diag-qtype=0", "--var-diag-qnum=1000",
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )
        # Variance diagnostic (Quadrature)
        outsig = call(["python", script_path, "--input=" + self.tm_fname,
                       "--var-diag=exact-base",
                       "--var-diag-qtype=3",
                       "--var-diag-qnum=%s" % ','.join(['3']*self.dim),
                       "--output=" + self.post_fname])
        self.assertFalse( outsig )

    def test_laplace(self):
        from distutils.spawn import find_executable
        from subprocess import call
        # Parameters
        tol = 1e-3
        ders = 2
        self.tm_fname = 'Laplace.dill'
        self.fname_list.append( self.tm_fname )
        # Find path to script (to avoid long paths with sh..)
        script_path = find_executable("tmap-laplace")
        # Run laplace script
        outsig = call(["python", script_path, "--dist=" + self.dist_fname,
                       "--output=" + self.tm_fname])
        self.assertFalse( outsig )
        # Test post-process
        self.postprocess()

    def test_direct(self):
        from distutils.spawn import find_executable
        from subprocess import call
        # Parameters
        tol = 1e-3
        ders = 2
        self.tm_fname = 'Direct.dill'
        self.fname_list.append( self.tm_fname )
        # Find path to script (to avoid long paths with sh..)
        script_path = find_executable("tmap-tm")
        # Run direct script
        outsig = call(["python", script_path, "--dist=" + self.dist_fname,
                       "--output=" + self.tm_fname,
                       "--mtype=intexp", "--span=total", "--btype=poly", "--order=1",
                       "--qtype=0", "--qnum=1000"])
        self.assertFalse( outsig )
        # Test post-process
        self.postprocess()

    def test_direct_xml(self):
        from distutils.spawn import find_executable
        import os.path
        from subprocess import call
        # Parameters
        tol = 1e-3
        ders = 2
        self.tm_fname = 'Direct.dill'
        self.fname_list.append( self.tm_fname )
        # Find path to script (to avoid long paths with sh..)
        script_path = find_executable("tmap-tm")
        # Run direct script
        test_map_xml_dir = os.path.dirname(os.path.realpath(__file__)) + \
                           '/xml/maps/'
        xml_map_fname = 'TotalOrdIntExpLinearMap_5d.xml'
        outsig = call(["python", script_path, "--dist=" + self.dist_fname,
                       "--output=" + self.tm_fname,
                       "--map-descr=" + test_map_xml_dir + xml_map_fname,
                       "--qtype=0", "--qnum=1000"])
        self.assertFalse( outsig )
        # Test post-process
        self.postprocess()

def build_suite(ttype='all'):
    suites_list = []
    if ttype in ['all','serial']:
        scripts_suite = unittest.TestLoader().loadTestsFromTestCase( Scripts_unittest )
        suites_list.append( scripts_suite )
    all_suites = unittest.TestSuite(suites_list)
    return all_suites

def run_tests(ttype='all'):
    all_suites = build_suite(ttype)
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()
